#
# wcl.tcl - a tcl clone of parts of the WCL (widget creation) library
# (not all wcl functionality is available, and some parts of it are made
#  quite specific to TK, I'm afraid...)
#
# There is one important difference between the widget 'paths' that
# we use here, and the standard TK widget paths:
# in TK, a menu is a widget, but a menu-entry is not a widget on its own.
# In this library, we simply 'act' as if an entry is just a subwidget
# (child) of the menu, so, for example, the path of a quit button
# will be something like '...menu.quit'
# When needed, we can use the utility routine wcWidgetOrEntry on a 'path'
# which returns either
# -  one item, the path itself, or
# -  a list of two items: path-of-the-menu subpath-of-entry
# The subpath-of-entry can be translated into a menu-index through
# the global wcl variable.
#
# Name convention:
# routines whose name start with 'Wc' are 'public' routines.
# the others (with names that start with 'wc' are private,
# not to be used by other modules.
#

# =========================================================================
#
# library routine loading / argument processing / config file reading
#

# =============================================================
#
# Interface routines (exported)
#
#

proc WcInit {} {
	global wcl
	set wcl(acctomake) {}
	set wcl(focusRedirectedFrom) {}
	set wcl(actioncount) 0
	set wcl(dummyFrame) .wcInternalDummy
	set wcl(windows) {}
	set wcl(vars,) {}
	set wcl(pane,sashwidth) 0
	set wcl(pane,sashhighlightbackground) \#ececec
	set wcl(pane,sashbackground) \#d9d9d9
	set wcl(pane,linewidth) 2
	set wcl(pane,linecolor) red
	set wcl(pane,opaquepaneresize) 0
	set wcl(pane,opaquepaneresize,showline) 0
	set wcl(cmdstofinalize) {}
}

proc WcOpaquePaneResize {{v ""}} {
	global wcl
	set old $wcl(pane,opaquepaneresize)
	if {![string match "" $v]} {
		set wcl(pane,opaquepaneresize) $v
	}
	return $old
}

proc WcExit { window {e {}} {clientData 0} {ignored {}}  } {
	if {$clientData == {}} {
		set clientData 0
	} elseif {![regexp -nocase {^-?[0-9]+} $clientData match]} {
		set clientData -1
	}
	exit $clientData
}

# arguments:
# app is used for a lot of things.
# class and resourceFileName are only used to find/read the resource
# file that describes the window/widget hierarchy.
proc WcCreate {{app main} {class {}} {resourceFileName {}}} {
	WcInit
	Debug_Msg WcCreate   "wcCreate1 current: [winfo children .]"
	wcReadResourceFile $app $class $resourceFileName
#	if {![catch {info body wcHookFontSetup}]} {
#		wcHookFontSetup
#	}
	catch { wcHookFontSetup }
	Debug_Msg WcCreate "wcCreate2 current: [winfo children .]"
	wcSetProtocolHandlers . $app
	Debug_Msg WcCreate   "wcCreate3 current: [winfo children .]"
	wcSetTitle  . $app
	Debug_Msg WcCreate   "wcCreate4 current: [winfo children .]"
	wcCreateChildren . $app
	Debug_Msg WcCreate   "wcCreate5 current: [winfo children .]"
	wcSetMenubar . $app
	Debug_Msg WcCreate   "wcCreate6 current: [winfo children .]"
	wcCreateAccelerators
	Debug_Msg WcCreate   "wcCreate7 current: [winfo children .]"
	wcCreateFocusRedirects
	Debug_Msg WcCreate   "wcCreate8 current: [winfo children .]"
	wcExecuteInitCallback . $app
	Debug_Msg WcCreate   "wcCreate9 current: [winfo children .]"

	update ;# we _need_ this,
	update idletasks ;# we _need_ this,
	# to be sure that the window sizes are known!!!
	# It has to be update; update idletasks is not enough,
	# at least on cygwin
	# but also plain update is not enough on suns, sometimes
	wcExecCreateFinalisationCmds
	Debug_Msg WcCreate   "wcCreate10 current: [winfo children .]"
}

proc wcExecuteInitCallback { w {name ""} } {
	global wcl
	if {$name != ""} {
		if {$w == "."} {
			set path .$name
		} else {
			set path $w.$name
		}
	} else {
		set path $w
	}
	set cb [wcGetResource $path initCallback]
	if {$cb != ""} {
		set wcl(cb,init,$w) [wcParseCallback $cb]
		wcExecCB $w init
	}
}

proc WcFullNameToWidget { w s } {
	global wcl
	if {$s == "."} {
		set result $s
	} else {
		set firstmatch [lsearch -glob $wcl(windows) $w$s]
		if {$firstmatch == -1} {
			set firstmatch [lsearch -glob $wcl(windows) $s]
		}
		if {$firstmatch == -1} {
			wcError WcFullNameToWidget "no widget matching $w $s"
			set result ""
		} else {
			set result [lindex $wcl(windows) $firstmatch]
		}
	}
	Debug_Msg "WcFullNameToWidget" "$w $s: $result"
	return $result
}

proc WcSetInSensitive { w e  windowlist {ignored {}} } {
	wcSetSensitivity $windowlist disabled
}
proc WcSetSensitive { w e windowlist {ignored {}} } {
	wcSetSensitivity $windowlist normal
}
proc WcSetCondInSensitive { w e  windowlist {ignored {}} } {
	wcSetCondSensitivity $windowlist disabled
}
proc WcSetCondSensitive { w e windowlist {ignored {}} } {
	wcSetCondSensitivity $windowlist normal
}


proc WcButtonGetSetState { widget {v -1}} {
	global wcl
	set result 0
	set b [wcWidgetOrEntry $widget]
	if {[llength $b] == 2} {
		set w [lindex $b 0]
		set i [lindex $b 1]
		set var $wcl(menuvars,$w,$i)
	} else {
		set w [lindex $b 0]
		set var $wcl(vars,$w)
	}
	if [info exists $var] {
		if {$v != -1} {
			# not the default value, so set it!
			set $var $v
		}
		set result [set $var]
	}
	return $result
}
proc WcButtonGetVarname { widget } {
	global wcl
	set result ""
	set b [wcWidgetOrEntry $widget]
	if {[llength $b] == 2} {
		set w [lindex $b 0]
		set i [lindex $b 1]
		set var $wcl(menuvars,$w,$i)
	} else {
		set w [lindex $b 0]
		set var $wcl(vars,$w)
	}
	if [info exists $var] {
		set result $var
	}
	return $result
}

proc WcSetLabel { widget s } {
	global wcl
	set l [wcWidgetOrEntry $widget]
	if {[llength $l] == 2} {
		set w [lindex $l 0]
		set i [lindex $l 1]
		if [winfo exists $w] {
			$w entryconfigure $wcl(menuindex,$w,$i) -label $s
		}
	} else {
		set w [lindex $l 0]
		if [winfo exists $w] {
			$w configure -text $s
		}
	}
}

proc WcGetLabel { widget } {
	global wcl
	set result ""
	set l [wcWidgetOrEntry $widget]
	if {[llength $l] == 2} {
		set w [lindex $l 0]
		set i [lindex $l 1]
		if [winfo exists $w] {
			set result [lindex [$w entryconfigure $wcl(menuindex,$w,$i) -label] 4]
		}
	} else {
		set w [lindex $l 0]
		if [winfo exists $w] {
			set result [lindex [$w configure -text] 4]
		}
	}
	return $result
}

proc WcSetCommandPrompt { widget s } {
	global wcl
	set wcl(prompt,$widget) $s
}

proc WcWindows {} {
	global wcl
	return $wcl(windows)
}

proc WcSetCursor {} {
	global wcl
	wcDoSetCursor .
	foreach w $wcl(windows) {
			wcDoSetCursor $w
	}
}
proc WcResetCursor {} {
	global wcl
	wcDoResetCursor .
	foreach w $wcl(windows) {
			wcDoResetCursor $w
	}
}

proc WcUnmanageCB { window {e {}} {w {}} {geo ok} } {
	WcUnmanage $window $e $w $geo
}
proc WcUnmanage { window {e {}} {windowString {}} {geo ok} } {
	global wcl
	regsub -all "," $windowString " " windows
	foreach w $windows {
		set path [WcFullNameToWidget "." $w]
		if {$path != ""} {
			wcDoUnmanage $path $geo
		}
	}
}

proc WcManageCB { window {e {}} {w {}} {ignored {}} } {
	WcManage $window $e $w
}
proc WcManage { window {e {}} {windowString {}} {ignored {}} } {
    global wcl
	Debug_FnArgs WcManage $window $e $windowString
	regsub -all "," $windowString " " windows
	foreach w $windows {
		set path [WcFullNameToWidget "." $w]
		if {$path != ""} {
			wcDoManage $path
		}
	}
}

proc WcSetFocus { window {e {}} w {ignored {}} } {
	global wcl
	set path [WcFullNameToWidget "." $w]
	if [winfo exists $w] {
		focus $w
	}
}

proc WcSelectLinesCB { window {e {}} {w {}} pos } {
	global wcl
	set start "[charsToIndexS $pos] linestart"
	set end "[charsToIndexS $pos] lineend"
	if [winfo exists $window] {
		selection clear $window
		$window tag add sel $start $end
	}
}

proc WcSystem { window {e {}} {command {}} ignored } {
	#foreach c [concat exec $command >& /dev/null] {
	#	lappend com $c
	#}
	#eval [concat $com]
	# we fist had 'concat' here below, instead of list,
	# but then we lost arguments in the invocation of
	# the command.
	# eval [list exec sh -c $command]
	# The above keeps tghe main process waiting for the
	# executed one to return, even if the $command contains
	# a & at the end. The one below works better, or so it
	# seems...
	eval exec $command
}


# =============================================================
# 
# Utility routines
# (for local use, not exported!)
#

proc wcGetPosition { args } {
	set result ""
	eval [concat Debug_FnArgs "wcGetPosition" $args]
	if {[llength $args] == 2} {
		set w [lindex $args 0]
		set p [lindex $args 1]
		if {[winfo exists $w]} {
			set result [rangeToChars $w [$w index $p]]
		}
	}
	Debug_FnRes "wcGetPosition" $result
	return $result
}

# ============================

proc wcWidgetOrEntry { w } {
	set result ""
	if {[winfo exists $w]} {
		set result [list $w]
	} else {
		set ldot [string last "." $w]
		if {$ldot >= 0} {
			incr ldot -1
			set mw [string range $w  0 $ldot]
			incr ldot 2
			set mi [string range $w $ldot end]
			set result [list $mw $mi]
		}
	}
	return $result
}

# ============================

proc capitalize { s } {
	return "[string toupper [string range $s 0 0]][string range $s 1 end]"
}

# ============================
#
# Callback handling:
# this routine invokes the 'real' callbacks as specified
# in the resource file.
# This routine is 'installed' as 'command' in the widgets

# general handler: invoke the callbacks that were defined as 'type'
# on widget 'w'.
# for the 'motion' type, we expect that args contains the 'x' and 'y'
# of the mouse event (click or motion)
proc wcExecCB { w type args } {
	global wcl
	eval [concat Debug_FnArgs "wcExecCB" $w $type $args]
#	if {[info exists wcl(cb-busy,$type,$w)] && $wcl(cb-busy,$type,$w)} {
#		Debug_Msg "wcExecCB" "$type locked for $w"
#		return
#	} else {
#		Debug_Msg "wcExecCB" "$type locking for $w"
#	}
#	set wcl(cb-busy,$type,$w) 1
	set cb $wcl(cb,$type,$w)
	# the calldata can be callback-type specific
	set calldata {}
	switch -- $type {
		motion {
			if {[llength $args] == 2} {
				set x [lindex $args 0]
				set y [lindex $args 1]
				set calldata [wcGetPosition $w @$x,$y]
			}
		}
		input {
			if {[llength $args] == 1} {
				set calldata [lindex $args 0]
			}
		}
		select {
			# This code is here for the listbox widget.
			# Check if there is also a 'deselect' callback for the widget, and if so,
			# see if we should 'set it up' to be invoked if the widget looses the
			# selection.
			if {[info exists wcl(cb,deselect,$w)]} {
				if {![catch {set selexported [lindex [$w configure -exportselection] 4] } errMsg]} {
					if {$selexported} {
						# NOTE: we NEED the after 500; without it this code does not work;
						# Most likely the problem is that there somehow is interference between
						# what WE do with the selection, and what the tk widgets do with it...
						# NOTE: moving this code to _after_ the invocation of the callbacks
						# (i.e. after the 'foreach cbe' below) does NOT solve that problem.
						after 500 \
							" Debug_Msg \"wcExecCB\"  \"selection own -command \\\"wcExecCB $w deselect\\\" $w\"
							  selection own -command \"wcExecCB $w deselect\" $w
							  Debug_Msg \"wcExecCB\"  \"selection: \[selection own\]\"
							"
					} else {
						Debug_Msg "wcExecCB"  "selection not exported"
					}
				} else {
					Debug_Msg "wcExecCB"  "Could not get $w configure -exportselection"
				}
			} else {
				Debug_Msg "wcExecCB"  "does not exist: wcl(cb,deselect,$w)"
			}
		}
		deselect {
			# This code is here for the listbox widget,
			# to explicitly deselect everything when we loose the
			# selection.. Do we really want to do this???
			# NOTE: we only do this IF there is a deselect callback present.
			# HMM.. should we really check that (as we do right now, in the next line?)
			if {[info exists wcl(cb,$type,$w)]} {
				# catch {$w selection clear}
				catch {$w selection clear 0 end}
			}
		}
	}
	foreach cbe $cb {
		set fn [lindex $cbe 0]
		set params [lindex $cbe 1]
		if {[catch {info args $fn}]} {
			if {![auto_load $fn]} {
				wcWarning "autoloading of proc failed (for $w): $fn"
			}
		}
		if {[lsearch [info procs] $fn] >= 0} {
			set com [list $fn]
			switch -glob -- $type {
				"actions*" {
					lappend com $w "event"
					set com [concat $com $params]
				}
				default {
					lappend com $w "event" $params
					# if {$calldata != {}} {
						lappend com $calldata
					# }
				}
			}
			Debug_Msg "wcExecCB" "$type calling for $w: $com"
			eval $com
		} else {
			wcWarning "unknown callback (for $w) ignored: $fn"
		}
	}
	Debug_Msg "wcExecCB" "$type freeing for $w"
#	set wcl(cb-busy,$type,$w) 0
}

#
# parse a callback string. return a list of pairs: (callback, parameters)
#
proc wcParseCallback { cb } {
	set callbacks {}
	set cbtemp [string trim $cb]
	while {[string length $cbtemp] > 0} {
		# puts "inwhile $cbtemp"
		if {[regexp -nocase {^([-0-9a-z_]+)[ 	]*\((.*)} $cbtemp match name rest]} {
			set i 0
			set goOn 1
			set parcount 1
			set paramend 0
			while {($parcount > 0) && ($i < [string length $rest])} {
				# puts "in2while $i $rest"
				switch -- [string range $rest $i $i] {
					"(" { incr parcount }
					")" { incr parcount -1 }
				}
				incr i
			}
			set j [expr $i - 1]
			if {[string range $rest $j $j] == ")"} {
				set paramend [expr $j - 1]
					lappend callbacks [list $name [string range $rest 0 $paramend]]
			}
			set cbtemp [string trimleft [string range $rest [expr $i] end]]
		} elseif {[regexp -nocase {^([-0-9a-z_]+)(.*)} $cbtemp match name rest]} {
			lappend callbacks [list $name {}]
			set cbtemp [string trimleft $rest]
		} else {
			wcError "wcParseCallback" "match failed for ($cbtemp)"
			set cbtemp ""
		}
	}
	return $callbacks
}


# ============================

# get title for window from resources, and set it

proc wcSetTitle { w {name ""} } {
	global wcl
	if {$name != ""} {
		if {$w == "."} {
			set path .$name
		} else {
			set path $w.$name
		}
	} else {
		set path $w
	}
	set title [wcGetResource $path title]
	if {$title != ""} {
		wm title $w $title
	}
}

# ============================

# get menu for window from resources, and set it

proc wcSetMenubar { w {name ""} } {
	global wcl
	if {$name != ""} {
		if {$w == "."} {
			set path .$name
		} else {
			set path $w.$name
		}
	} else {
		set path $w
	}
	set menu [wcGetResource $path wcMenu]
	if {$menu != ""} {
		Debug_Msg wcCreateChild "$w configure -menu .$menu"
		$w configure -menu .$menu
	}
}

# ============================


# get protocol handler callbacks from resources, and register them
# we use the name argument (if given) to adapt the name that we use
# to find the resources

proc wcSetProtocolHandlers { w {name ""} } {
	global wcl
	if {$name != ""} {
		if {$w == "."} {
			set path .$name
		} else {
			set path $w.$name
		}
	} else {
		set path $w
	}
	set cb [wcGetResource $path deleteWindowCallback]
	if {$cb != ""} {
		set wcl(cb,deleteWindow,$w) [wcParseCallback $cb]
		wm protocol $w WM_DELETE_WINDOW "wcExecCB $w deleteWindow"
	}
	set cb [wcGetResource $path saveYourselfCallback]
	if {$cb != ""} {
		set wcl(cb,saveYourself,$w) [wcParseCallback $cb]
		wm protocol $w WM_SAVE_YOURSELF "wcExecCB $w saveYourself"
	}
}

# ============================

# get help callbacks from resources, and register them
# we use the name argument (if given) to adapt the name that we use
# to find the resources

proc wcContextSensitiveHelp { w {name ""} } {
	global wcl
	if {[winfo exists $w]} {
		if {$name != ""} {
			if {$w == "."} {
				set path .$name
			} else {
				set path $w.$name
			}
		} else {
			set path $w
		}
		set cb [wcGetResource $path helpCallback]
		if {$cb != ""} {
			set wcl(cb,help,$w) [wcParseCallback $cb]
			set event [wcParseEvent "<<Help>>"]
			bind $w $event "wcExecCB $w help"
		}
	}
}

# ============================

# remove the firs occurence of item i from list l (use 'exact' matching)
# (duplicate occurences are not removed)
proc wcRemoveListItem { i l } {
	set res $l
	set found [lsearch -exact $l $i]
	if {$found != -1} {
		set res [lreplace $l $found $found]
	}
	return $res
}

# ============================
#
# widget creation.
# we do this in a number of recursive routines, that call each other
#

# create the children of widget w (and their children, and...)
# uses wcCreateChild to create each child
proc wcCreateChildren {w {name {}}} {
	if [winfo exists $w] {
		if {$name == {}} {
			set childrenString [wcGetResource $w wcChildren]
			set menuString [wcGetResource $w wcMenu]
		} else {
			set dummy [wcCreateDummy $w $name]
			set childrenString [wcGetResource $dummy wcChildren]
			set menuString [wcGetResource $dummy wcMenu]
			wcDestroyDummy $dummy
		}
		regsub -all "," $childrenString " " children
		regsub -all "," $menuString " " menu
		Debug_Msg "wcCreateChildren" "Yes $w $name ($children)($menu)"
		foreach child $children {
			wcCreateChild $w $child
		}
		foreach child $menu {
			wcCreateChild $w $child
		}
	} else {
		Debug_Msg "wcCreateChildren" "No $w $name"
	}
}

# create a child of widget w, and its children
# this routine is mainly a switch over the widget 'class', in which for
# each class the 'correct' routine is used to create the widget,
# after which widget specific resources are automatically)
# uses wcCreateChildren to create the children
proc wcCreateChild {w child } {
	global wcl
	Debug_FnArgs "wcCreateChild" $w $child
	###puts  "wcCreateChild $w $child"
	###puts  "wcCreateChild current: [winfo children .]"
	set name {}
	if {$w == "."} {
		set childpath .$child
	} else {
		set childpath $w.$child
	}
	if ![winfo exists $childpath] {
		lappend wcl(windows) $childpath
		set class [wcGetResourceDummy $w $child class]
		set create [wcGetResourceDummy $w $child wcCreate]
		set pack [wcGetResourceDummy $w $child pack]
		set menu [wcGetResourceDummy $w $child wcMenu]
		set label [wcGetResourceDummy $w $child label]
		set command [wcGetResourceDummy $w $child command]
		set isset [wcGetResourceDummy $w $child set]
		set variable [wcGetResourceDummy $w $child variable]
		set sensitive [wcGetResourceDummy $w $child sensitive]
		Debug_Msg wcCreateChild "$childpath ($w)($child)($childpath)($class)($create)($pack)($menu)($label)($command)($isset)"
		# frame $childpath
		set params [lrange $create 1 end]
		set doNotCreateChildren 0
		if {$create != ""} {
			switch -- [lindex $create 0] {
				Toplevel {
					set c Widget_Toplevel
					set res [$c $childpath $child $class]
					set man [wcGetResource $w.$child wcManaged]
					if {$man != ""} {
						switch -- [string tolower $isset] {
							true -
							on -
							1 {
								# do nothing
							}
							default {
								# unmanage
								wm withdraw $res
							}
						}
					}
					wcSetTitle $res
					wcSetMenubar $res
					wcSetProtocolHandlers $res
				}
				Frame {
					set c Widget_Frame
					lappend com $c $w $child $class
					if {$pack != ""} {
						lappend com $pack
					} else {
						lappend com {top expand fill}
					}
					Debug_Msg wcCreateChild-Frame [concat $com $params]
					eval [concat $com $params]
				}
				VPane {
					set c Widget_Frame
					lappend com $c $w $child $class
					lappend params -bg black
					if {$pack != ""} {
						lappend com $pack
					} else {
						lappend com {top expand fill}
					}
					Debug_Msg wcCreateChild-VPane [concat $com $params]
					eval [concat $com $params]
					wcAddFinalisationCmd [list wcMakeVPane $w $childpath]
				}
				HPane {
					set c Widget_Frame
					lappend com $c $w $child $class
					lappend params -bg black
					if {$pack != ""} {
						lappend com $pack
					} else {
						lappend com {top expand fill}
					}
					Debug_Msg wcCreateChild-HPane [concat $com $params]
					eval [concat $com $params]
					wcAddFinalisationCmd [list wcMakeHPane $w $childpath]
				}
				Label {
					set c Widget_Label
					lappend com $c $w $child
					if {$pack != ""} {
						lappend com $pack
					} else {
						lappend com {left fill}
					}
					eval [concat $com $params]
				}
				Menu {
					set c Widget_Menu
					$c $w $child
					set title [wcGetResource $w.$child title]
					if {$title != ""} {
						$w configure -title $title
					}
				}
				MenuButton {
					set c Widget_AddMenuBDefNoMenu
					if {$pack != ""} {
						set res [$c $w $child $pack]
					} else {
						set res [$c $w $child]
					}
					Debug_Msg wcCreateChild "$res configure -menu $res.$menu"
					$res configure -menu $res.$menu
				}
				Button {
					set c Widget_AddBut
					if {$pack != ""} {
						set res [$c $w $child "" "" $pack]
					} else {
						set res [$c $w $child]
					}
					set cb [wcGetResource $w.$child activateCallback]
					if {$cb != ""} {
						set wcl(cb,activate,$res) [wcParseCallback $cb]
						$res configure -command "wcExecCB $res activate"
					}
				}
				RadioButton {
					set c Widget_RadioBut
					set wcl(menuvar,$w,$child) wcl(var,$variable)
					if {$pack != ""} {
						set res [$c $w $child $label $wcl(menuvar,$w,$child) $pack]
					} else {
						set res [$c $w $child $label $wcl(menuvar,$w,$child)]
					}
					set cb [wcGetResource $w.$child activateCallback]
					if {$cb != ""} {
						set wcl(cb,activate,$res) [wcParseCallback $cb]
						$res configure -command "wcExecCB $res activate"
					}
					if {$isset != ""} {
						set var $wcl(menuvar,$w,$child)
						if {[info exists $var]} {
							switch -- [string tolower $isset] {
								true -
								on -
								1 {
									set $var $label
								}
							}
						}
						Debug_Msg wcCreateChild "$c $var [set $var]"
					}
				}
				CheckButton {
					set c Widget_CheckBut
					set wcl(menuvar,$w,$child) wcl(var,$variable)
					if {$pack != ""} {
						set res [$c $w $child $label $wcl(menuvar,$w,$child) $pack]
					} else {
						set res [$c $w $child $label $wcl(menuvar,$w,$child)]
					}
					set cb [wcGetResource $w.$child activateCallback]
					if {$cb != ""} {
						set wcl(cb,activate,$res) [wcParseCallback $cb]
						$res configure -command "wcExecCB $res activate"
					}
					if {$isset != ""} {
						set var $wcl(menuvar,$w,$child)
						if {[info exists $var]} {
							switch -- [string tolower $isset] {
								true -
								on -
								1 {
									set $var 1
								}
							}
						}
						Debug_Msg wcCreateChild "$c $var [set $var]"
					}
				}
				ButtonDef {
					set c Widget_AddButDef
					$create $w $child
				}
				ButtonItem {
					set c Widget_AddMenuItem
					$c $w $label $command
					lappend wcl(menuentries,$w) $child
					lappend wcl(menuindex,$w,$child) [$w index $label]
					lappend wcl(menulabels,$w,$child) $label
					set cb [wcGetResource $w.$child activateCallback]
					if {$cb != ""} {
						set wcl(cb,activate,$w.$child) [wcParseCallback $cb]
						$w entryconfigure $label -command "wcExecCB $w.$child activate"
					}
					set acc [wcGetResource $w.$child acceleratorText]
					if {$acc != ""} {
						set accBinding [wcParseEvent $acc "unknown"]
						if {$accBinding != "unknown"} {
							$w entryconfigure $label -accelerator $acc
							wcMakeAcceleratorBinding $w $accBinding [list $w invoke $label]
						} else {
							wcWarning "unknown binding: $acc"
						}
					}
					set underline [wcGetResource $w.$child underline]
					if {$underline != ""} {
						$w entryconfigure $label -underline $underline
					}
				}
				CheckItem {
					set c Widget_CheckMenuItem
					set res [$c $w $label $command wcl(menuvar,$w,$child)]
					lappend wcl(menuentries,$w) $child
					lappend wcl(menuindex,$w,$child) [$w index $label]
					lappend wcl(menulabels,$w,$child) $label
					lappend wcl(menuvars,$w,$child) [$w entrycget $label -variable]
					set cb [wcGetResource $w.$child valueChangedCallback]
					if {$cb != ""} {
						set wcl(cb,valueChanged,$w.$child) [wcParseCallback $cb]
						$w entryconfigure $label -command "wcExecCB $w.$child valueChanged"
					}
					if {$isset != ""} {
						set var $wcl(menuvars,$w,$child)
						if [info exists $var] {
							switch -- [string tolower $isset] {
								true -
								on -
								1 {
									set $var 1
								}
								default {
									set $var 0
								}
							}
						}
						Debug_Msg wcCreateChild "$c $var [set $var]"
					}
				}
				RadioItem {
					set c Widget_RadioMenuItem
					set wcl(menuvar,$w,$child) wcl(var,$variable)
					set res [$c $w $label $wcl(menuvar,$w,$child)]
					lappend wcl(menuentries,$w) $child
					lappend wcl(menuindex,$w,$child) [$w index $label]
					lappend wcl(menulabels,$w,$child) $label
					lappend wcl(menuvars,$w,$child) [$w entrycget $label -variable]
					set cb [wcGetResource $w.$child activateCallback]
					if {$cb != ""} {
						set wcl(cb,valueChanged,$w.$child) [wcParseCallback $cb]
						$w entryconfigure $label -command "wcExecCB $w.$child valueChanged"
					}
					if {$isset != ""} {
						set var $wcl(menuvar,$w,$child)
						if {[info exists $var]} {
							switch -- [string tolower $isset] {
								true -
								on -
								1 {
									set $var $label
								}
							}
							Debug_Msg wcCreateChild "$c $var [set $var]"
						} else {
							Debug_Msg wcCreateChild "$c $var"
						}
					}
				}
				SeperatorItem {
					set c Widget_AddMenuSeparator
					$c $w
				}
				CascadeItemWithMenu {
					set c Widget_CascadeMenuItem
					$c $w $label $command $menu
					lappend wcl(menuentries,$w) $child
					lappend wcl(menuindex,$w,$child) [$w index $label]
					lappend wcl(menulabels,$w,$child) $label
				}
				CascadeItem {
					set c Widget_CascadeMenuItemNoMenu
					wcCreateChild $w $menu
					$c $w $label $command $w.$menu
					lappend wcl(menuentries,$w) $child
					lappend wcl(menuindex,$w,$child) [$w index $label]
					lappend wcl(menulabels,$w,$child) $label
					set doNotCreateChildren 1
					set acc [wcGetResource $w.$child acceleratorText]
					if {$acc != ""} {
						set accBinding [wcParseEvent $acc "unknown"]
						if {$accBinding != "unknown"} {
							$w entryconfigure $label -accelerator $acc
							wcMakeAcceleratorBinding $w $accBinding [list $w invoke $label]
						} else {
							wcWarning "unknown binding: $acc"
						}
					}
					set underline [wcGetResource $w.$child underline]
					if {$underline != ""} {
						$w entryconfigure $label -underline $underline
					}
				}
				ScrolledText {
					set c Widget_Text
					lappend com $c $w $child 1 -wrap none
					eval [concat $com $params]
					set cb [wcGetResource $w.$child motionVerifyCallback]
					if {$cb != ""} {
						set wcl(cb,motion,$w.$child) [wcParseCallback $cb]
						bind $w.$child <Button1-Motion> { wcExecCB %W motion %x %y }
						bind $w.$child <ButtonPress-1> { wcExecCB %W motion %x %y }
						bindtags $w.$child [wcRemoveListItem Text [bindtags $w.$child]]
					}
					set editable [wcGetResource $w.$child editable]
					switch -- [string tolower $editable] {
						false -
						off -
						0 {
							$w.$child configure -state disabled
						}
					}
				}
				ScrolledListbox {
					set c Widget_Listbox
					lappend com $c $w $child wcl(listboxvar,$w,$child)
					eval [concat $com $params]
					set cb [wcGetResource $w.$child selectCallback]
					if {$cb != ""} {
						set wcl(cb,select,$w.$child) [wcParseCallback $cb]
						bind $w.$child <Button1-Motion> { wcExecCB %W select }
						bind $w.$child <ButtonPress-1> { wcExecCB %W select }
					}
					set cb [wcGetResource $w.$child deselectCallback]
					if {$cb != ""} {
						set wcl(cb,deselect,$w.$child) [wcParseCallback $cb]
					}
					set cb [wcGetResource $w.$child activateCallback]
					if {$cb != ""} {
						set wcl(cb,activate,$w.$child) [wcParseCallback $cb]
						bind $w.$child <Double-ButtonPress-1> { wcExecCB %W activate }
						bind $w.$child <Triple-ButtonPress-1> {}
					}
				}
				ScrolledCli {
					set c Widget_Text
					lappend com $c $w $child 1 -wrap none
					eval [concat $com $params]
					set prompt [wcGetResource $w.$child prompt]
					set wcl(prompt,$w.$child) $prompt
					$w.$child insert end $wcl(prompt,$w.$child)
					$w.$child mark set promptend "insert lineend"
					$w.$child mark gravity promptend left
					bindtags $w.$child [wcRemoveListItem Text [bindtags $w.$child]]
					set mycode  {
						if {![wcGT %W insert promptend]} {
							%W mark set insert promptend
						}
					}
					foreach ev [bind Text] {
						set defcode [bind Text $ev]
						bind $w.$child $ev "$defcode $mycode"
					}
					set cb [wcGetResource $w.$child inputCallback]
					if {$cb != ""} {
						set wcl(cb,input,$w.$child) [wcParseCallback $cb]
						bind $w.$child <Return> {
							%W insert end \n
							set line [%W get promptend "promptend lineend"]
							%W mark set insert end
							if {$line == {}} {
								set line ""
							}
							Debug_Msg wcCreateChild-ScrolledCli "%W $line"
							wcExecCB %W input $line
							%W insert end $wcl(prompt,%W)
							catch {%W mark unset promptend}
							%W mark set promptend "insert lineend"
							%W mark gravity promptend left
							%W see promptend
						}
					}
					foreach ev {<BackSpace> <Meta-BackSpace>} {
						set defcode [bind Text $ev]
						set testcode {
							if {[wcGT %W insert promptend]} {
								DODEF
							}
							MYCODE
						}
						regsub {DODEF} $testcode $defcode code1
						regsub {MYCODE} $code1 $mycode code
						bind $w.$child $ev $code
					}
					set editable [wcGetResource $w.$child editable]
					switch -- [string tolower $editable] {
						false -
						off -
						0 {
							$w.$child configure -state disabled
						}
					}
				}
				Entry {
					set c Widget_Entry
					lappend com $c $w $child
					if {$pack != ""} {
						lappend com $pack
					} else {
						lappend com {left fill}
					}
					eval [concat $com $params]
					set v [wcGetResource $w.$child value]
					if {$v != ""} {
						$w.$child insert 0 $v
					}
				}
				Sash {
					set v [wcGetResource $w.$child resize]
					set l [wcGetResource $w.$child errorLabel]
					if {($v != "") && ($l != "")} {
						set wv [WcFullNameToWidget "." $v]
						set wl [WcFullNameToWidget "." $l]
						ExwinFtocMsgBoundary $w $child $wv $wl
					}
				}
				default {
					$create $w $child
				}
			}
			if {$sensitive != ""} {
				switch -- [string trim [string tolower $sensitive]] {
					false -
					off -
					0 {
						wcSetCondSensitivityForWidget [wcWidgetOrEntry $w.$child] disabled
					}
					default {
					}
				}
			}
		} else {
		wcError wcCreateChild "no or empty wcCreate resource for $w $child"
			Debug_Msg wcCreateChild "$w $child ($create) ($class)"
		}
		if {! $doNotCreateChildren} {
			wcCreateChildren $childpath $name
		}
		set translations [wcGetResource $childpath translations]
		if {$translations != ""} {
			wcAddTranslations $childpath $translations
		}
		wcContextSensitiveHelp $childpath $name
	}
}

# ============================
#
# resource 'getting'
# we have two routines, one for widgets that already exist, and one
# for those that do not.
#
# NOTE: this will not find all resources for widgets that do not (yet) exist!!!
#
# For example, for widget .a.b.c.d it will only find resources that
# are specified as *d:resource and not resources that are specified
# with more context information, like *b*d:resource
#
# The reason for this is that TK only allows us to get resources for
# widgets (windows) that do exist. So, for widgets that do not yet exist,
# we create a simple dummy window, get its resources, and delete it
# again. Currently, for a child widget '.some.path.to.child', the dummy
# widget created is '.dummy.child' .
# We could probably solve this problem if we would instead create a dummy
# widget like '.dummy.some.path.to.child' (for the same example).
#

# get a resource, for a widget that has not been created yet,
# (eg. because we have to know this resource to create it,
#  or because it isn't really a widget we want to resource for, but
#  a menu entry
# )
# NOTE: this will not find all resources!!!
# See the comment in the start of this 'section'.
#
# currently, we ignore the 'parent' field, but we use it in the future
proc wcGetResourceDummy {parent child resourceName {resourceClass ""}} {
	Debug_FnArgs "wcGetResourceDummy" $parent $child $resourceName $resourceClass
	if {$resourceClass == ""} {
		set resourceClass [capitalize $resourceName]
	}
	set dummyChild [wcCreateDummy $parent $child]
	set resourceString [option get $dummyChild $resourceName $resourceClass]
	wcDestroyDummy $dummyChild
	return $resourceString
}

# get a resource for a widget
# if the widget doesn't exist (yet), we use wcGetResourceDummy
# NOTE: for widgets that do not yet exist, it will not find all resources!!!
# See the comment in the start of this 'section'.
#
proc wcGetResource {widget resourceName {resourceClass ""}} {
	Debug_FnArgs "wcGetResource" $widget $resourceName $resourceClass
	if {$resourceClass == ""} {
		set resourceClass [capitalize $resourceName]
	}
	set b [wcWidgetOrEntry $widget]
	if {[llength $b] == 2} {
		set w [lindex $b 0]
		set i [lindex $b 1]
		set resourceString [wcGetResourceDummy $w $i $resourceName $resourceClass]
	} else {
		set w [lindex $b 0]
		set resourceString [option get $w $resourceName $resourceClass]
	}
	return $resourceString
}

# create a dummy widget, to be able to get its resources
# currently, we only create the dummy child (result path: .dummy.child )
# we might prefer to (recursively) create the whole parent path as well
# (so we get result path: .dummy.parentpath.child )
proc wcCreateDummy {parent child} {
	global wcl
	if {([string compare $parent ""] != 0) &&
	    ([string compare $parent "."] != 0)} {
		set parent_parts [split $parent .]
		# take care of initial . at start of string -> causes empty element
		if {[string compare [lindex $parent_parts 0] ""] == 0} {
			set parent_parts [lrange $parent_parts 1 end]
		}
		foreach p $parent_parts {
			lappend parts $p
		}
	}
	if {[string compare $child ""] != 0} {
		lappend parts $child
	}
	if {![info exists parts]} {
		set parts {}
	}
	set win [join $parts .]
	Debug_Msg wcCreateDummy "($parent) ($child); ($parts): ($wcl(dummyFrame))($win)"
	if ![winfo exists .$win] {
#		set dummy  $wcl(dummyFrame).$win
		if ![winfo exists $wcl(dummyFrame)] {
			frame $wcl(dummyFrame)
		}
		set dummywin $wcl(dummyFrame)
		foreach w $parts {
			set dummywin $dummywin.$w
			if ![winfo exists $dummywin] {
				frame $dummywin
			}
		}
		set res $dummywin
	} else {
		set res .$win
	}
	return $res
}
# destroy the dummy widget
proc wcDestroyDummy {path} {
	global wcl
	# destroy $path
}


# ============================
#
# sensitivity handling
#
proc wcSetCondSensitivityForWidget { b v {var ""}} {
	global wcl
	set result 0
	set w ""
	### puts "wcSetCondSensitivityForWidget $b $v $var"
	Debug_FnArgs "wcSetCondSensitivityForWidget"  "$b" "$v" "$var"
	if {$var != ""} {
		if {![info exists wcl(condsensitivity,$b,$var)]} {
			lappend wcl(condsensitivity,vars,$b) $var
			lappend wcl(condsensitivity,widgets,$var) $b
		}
		Debug_Msg "wcSetCondSensitivityForWidget" "$b $v $var Names: $wcl(condsensitivity,vars,$b)"
		switch -- $v {
			  disabled {
			  	set wcl(condsensitivity,$b,$var) 0
			} normal {
			  	set wcl(condsensitivity,$b,$var) 1
			  	foreach n $wcl(condsensitivity,vars,$b) {
					Debug_Msg "wcSetCondSensitivityForWidget" "Cond-enable $b $v $var $n: $wcl(condsensitivity,$b,$n)"
			  		if {! $wcl(condsensitivity,$b,$n)} {
			  			return
			  		}
			  	}
			}
		}
	}
	if {[llength $b] == 2} {
		set w [lindex $b 0]
		set i [lindex $b 1]
		Debug_Msg "wcSetCondSensitivityForWidget" "$b $v $w $i $wcl(menuindex,$w,$i) $var: ==============="
		catch {$w entryconfigure $wcl(menuindex,$w,$i) -state $v}
		Debug_Msg "wcSetCondSensitivityForWidget" "[$w entryconfigure $wcl(menuindex,$w,$i)]"
	} elseif {[string match [winfo class [lindex $b 0]] "Menu"]} {
		set w [lindex $b 0]
		set i 0
		set l [$w index last]
		while {([string compare $l none] != 0) && $i <= $l} {
			Debug_Msg "wcSetCondSensitivityForWidget" "$b $v $w $i $var: ==============="
		
			catch {$w entryconfigure $i -state $v}
			Debug_Msg "wcSetCondSensitivityForWidget" "[$w entryconfigure $i]"
			incr i
		}
	} else {
		set w [lindex $b 0]
		Debug_Msg "wcSetCondSensitivityForWidget" "$b $v $var: ==============="
		catch {$w configure -state $v}
	}
	Debug_FnRes "wcSetCondSensitivityForWidget"
}
proc wcSetSensitivity { windowlist val } {
	regsub -all "," $windowlist " " windows
	foreach window $windows {
		wcSetCondSensitivityForWidget [wcWidgetOrEntry [WcFullNameToWidget "." $window]] $val
	}
}
proc wcSetCondSensitivity { var_windowlist val } {
	regsub -all "," $var_windowlist " " var_windows
	set var [lindex $var_windows 0]
	set windows [lrange $var_windows 1 end]
	foreach window $windows {
		wcSetCondSensitivityForWidget [wcWidgetOrEntry [WcFullNameToWidget "." $window]] $val $var
	}
}

# ============================
#
# Internal error and warning routines.
#
proc wcError { fn args } {
	set sep " "
	catch {puts stderr "Wcl error in $fn: [join $args $sep]"}
}
proc wcWarning { args } {
	set sep " "
	catch {puts stderr "Wcl warning: [join $args $sep]"}
}

# ============================
#
# Finalisationof Creation
#
# here we collect code that is to be executed at the end of the
# wc-create call
#

proc wcAddFinalisationCmd { c } {
	global wcl
	lappend wcl(cmdstofinalize) $c
}

proc wcExecCreateFinalisationCmds {} {
	global wcl
	foreach c $wcl(cmdstofinalize) {
		eval $c
	}
}

# ============================
#
# Accelerator handling.
#
# we first collect accelerators, and make them only once all
# windows have been created, so we are sure that wcFullNameToWidget
# will be able to find all windows specified
#

# this routine is called from wcCreateChild to 'register' an accelerator
# that is to be created
proc wcMakeAcceleratorBinding { w acc command } {
	global wcl
	lappend wcl(acctomake) [list $w $acc $command]
}

# this routine is called from WcCreate to really create those accelerators
# that have been registered by  wcMakeAcceleratorBinding
# AS SIDE-EFFECT it registers focus redirects (in wcDoMakeAcceleratorBinding),
# that will be 'implemented' by wcCreateFocusRedirects
proc wcCreateAccelerators {} {
	global wcl
	foreach acctodo $wcl(acctomake) {
		set w [lindex $acctodo 0]
		set acc [lindex $acctodo 1]
		set command [lindex $acctodo 2]
		wcDoMakeAcceleratorBinding $w $acc $command
	}
}

# this routine is called from wcCreateAccelerators
# it creates the bindings that implement the accelerators, and
# AS SIDE-EFFECT it registers focus redirects, that will be
# 'implemented' by wcCreateFocusRedirects
# BUG: these focus redirects make the bindings for all accwindows,
#      except for one: the focusTarget, superflous!!!
proc wcDoMakeAcceleratorBinding { w acc command } {
	global wcl
	foreach mapping [wcGetAcceleratorWindows $w] {
		set focusedWidget [lindex $mapping 0]
		set accwindow [lindex $mapping 1]
		Debug_Msg wcDoMakeAcceleratorBinding "$accwindow $acc $command"
		bind $accwindow $acc $command
		if {[lsearch -exact $wcl(focusRedirectedFrom) $accwindow] == -1} {
			lappend wcl(focusRedirectedFrom) $accwindow
			set wcl(focusRedirected,$accwindow) $focusedWidget
		}
	}
}

# this routine is called from wcDoMakeAcceleratorBinding to find the
# 'target' window to which the accelerators should be bound,
# and the 'source' windows for which the focus should be redirected to
# that 'target'.
# returns a list of pairs (target, source)
proc wcGetAcceleratorWindows { w } {
	Debug_FnArgs "wcGetAcceleratorWindows" $w
	set res {}
	if {[winfo exists $w]} {
		set accW [wcGetResource $w acceleratorWindows]
		set focusTarget [wcGetResource $w focusTarget]
		if {$accW != 0} {
			set res {}
			regsub -all "," $accW " " accwindows
			foreach accwindow $accwindows {
				lappend res [list [WcFullNameToWidget $w $focusTarget] [WcFullNameToWidget $w $accwindow]]
			}
		}
	}
	if {$res == {}} {
		if {$w == "."} {
			lappend res [list "." "."]
		} else {
			set res [wcGetAcceleratorWindows [winfo parent $w]]
		}
	}
	Debug_FnRes "wcGetAcceleratorWindows" $res
	return $res
}


# this routine is called from WcCreate to make bindings that redirect the
# focus to a 'target' window when the mouse enters a 'source' window,
# as registered by wcDoMakeAcceleratorBinding
proc wcCreateFocusRedirects {} {
	global wcl
	foreach redirect $wcl(focusRedirectedFrom) {
		set target $wcl(focusRedirected,$redirect)
		Debug_Msg wcCreateFocusRedirects "redirecting focus from $redirect to $target"
		bind $redirect <Enter> "focus $target"
		bind $redirect <Leave> "focus ."
	}
}

# this routine parses 'translations' and creates bindings for them
proc wcAddTranslations { w t } {
	global wcl
	Debug_FnArgs "wcAddTranslations" $w $t
	set tmp [string trim $t]
	if {[regexp -nocase {^(#[0-9a-z_]+)?[ 	]*(.*)} $tmp match name rest]} {
		regsub -all {\\n} $rest "\n" restSubst
		# Debug_Msg "wcAddTranslations" "restSubst ($restSubst)"
		set lines [split $restSubst \n]
		set count 0
		foreach line $lines {
			set def [split $line ':']
			if {[llength $def] == 2} {
				set event [wcParseEvent [lindex $def 0]]
				set actions [lindex $def 1]
				incr count
				set wcl(cb,actions$count,$w) [wcParseCallback $actions]
				Debug_Msg "wcAddTranslations" "$w $event $wcl(cb,actions$count,$w)"
				bind $w $event "wcExecCB $w actions$count"
			}
		}
		set class [winfo class $w]
		if {$class != {}} {
			bindtags $w [wcRemoveListItem $class [bindtags $w]]
		}
	}
}

proc wcParseEvent { e {def "unknown"}} {
	Debug_FnArgs "wcParseEvent" $e
	set res $def
	switch -glob -- [string trim [string tolower $e]] {
		"<enter>" { set res <Return> }
		"ctrl-*" {
			regsub "ctrl" [string tolower $e] "Control" subst
			set res "<$subst>"
		}
		"->" { set res <Right> }
		"break" { set res <Key-F21> }
		"f*" { set res <$e> }
		"<btn1down>" { set res "<ButtonPress-1>" }
		"<btn2down>" { set res "<ButtonPress-2>" }
		"<btn3down>" { set res "<ButtonPress-3>" }
		"<mouseenter>" { set res "<Enter>" }
		"<mouseleave>" { set res "<Leave>" }
		"<key-find>" { set res "<F19>" }
		"<<help>>" { set res "<Help>" }
		default { set res $e }
	}
	Debug_FnRes "wcParseEvent" $res
	return $res
}
# ============================
#
# Cursor handling.
#
# change the cursor to a specified one, or back to the default one.
# When we change the cursor to a specified one, we remember the current
# cursor.
# We keep a stack of cursors, to allow several contexts, i.e.
#  - we change the cursor first, because we do a long execution (push cursor)
#  - we move the mouse over an active part of the window,
#       cursor changes again, (push cursor)
#  - execution is finished, (pop cursor) (now we have the watch, not nice, but OK)
#  - move mouse out of active part of the window, (pop cursor) (initial cursor back)
#

proc wcDoSetCursor { w {cursor watch}} {
	global wcl
	Debug_FnArgs "wcDoSetCursor" $w $cursor
	if {[winfo exists $w]} {
		set cur [$w cget -cursor]
		if {[info exists wcl(defaultcursor,$w)]} {
			set wcl(defaultcursor,$w) [concat $cur $wcl(defaultcursor,$w)]
		} else {
			set wcl(defaultcursor,$w) [list $cur ]
		}
		Debug_Msg "wcDoSetCursor" "$w storing cursor ($cur) stack ($wcl(defaultcursor,$w))"
		$w configure -cursor $cursor
	}
}

# the keep 'flag' of wcDoResetCursor should be used if the 'stack'
# of cursors should not be changed, i.e. if wcDoResetCursor will be called
# more often than wcDoSetCursor for a window.
# For example, in the following scenario:
#	WcSetCursor                    ;# set watch cursor for all, except:
#	wcDoResetCursor .sub.sub keep  ;# reset the cursor for a button, still enabled
#   ...
#	WcResetCursor                  ;# reset cursor for all
#
proc wcDoResetCursor { w {keep ""}} {
	global wcl
	Debug_FnArgs "wcDoResetCursor" $w
	if {[winfo exists $w]} {
		set ww $w
		set def ""
		while {($def == "") && ($ww != "")} {
			if [info exists wcl(defaultcursor,$ww)] {
				set def [lindex $wcl(defaultcursor,$ww) 0]
				if {[string compare $w $ww] == 0} {
					if {$keep == ""} {
						set wcl(defaultcursor,$ww) [lrange $wcl(defaultcursor,$ww) 1 end]
					}
				}
				Debug_Msg "wcDoResetCursor" "$ww found cursor ($def) stack ($wcl(defaultcursor,$ww))"
			}
			set ww [winfo parent $ww]
		}
#		if {$def == ""} {
#			set def top_left_arrow
#		}
		Debug_Msg "wcDoResetCursor" "$w setting cursor ($def)"
		$w configure -cursor $def
	}
}

# ============================
#
# Window (un)managing
#
# map or unmap a window
#
proc wcDoUnmanage { path geo } {
	global wcl
	if [winfo exists $path] {
    	switch -- $geo {
		"ok" {
		    set wcl(geometry,$path) [wm geometry $path]
		}
		"nosize" {
		    set wcl(geometry,$path) [string trimleft [wm geometry $path] 0123456789x]
		}
		default {
	    	catch {unset wcl(geometry,$path)}
		
    	}
    	}
	    if [info exists exwin(geometry,$path)] {
			# Some window managers return geometry like
			# 80x24+-1152+10
			regsub -all {\+-} $wcl(geometry,$path) + wcl(geometry,$path)
	    }
		wm withdraw $path
	}
}

# first check if window still exists. If not, try to create it
# then, invoke wcDoReallyManage to do the real work.
proc wcDoManage { path } {
    global wcl
	set name ""
	set class ""
    if ![winfo exists $path] {
    	set parts [split $path .]
    	set child [lindex $parts end]
    	set parts [lreplace $parts end end]
    	set prefix [join $parts .]
    	set w .$prefix
    	wcCreateChild $w $child
    }
	return [wcDoReallyManage $path]
}
proc wcDoReallyManage { path } {
    global wcl
    if [catch {wm state $path} state] {
    	return 1
    } else {
		if {$state != "normal"} {
		    catch {
				wm geometry $path $wcl(geometry,$path)
				Exmh_Debug Exwin_Toplevel $path $wcl(geometry,$path)
	 	   }
			wm deiconify $path
		} else {
		    catch {raise $path}
		}
		return 0
    }
}

proc wcGetLineNr { pos } {
	set result 0
	if {[regexp {^([0-9]+)\.} $pos match linenr] == 1} {
		set result $linenr
	}
	return $result
}
proc wcGetPosNr { pos } {
	set result 0
	if {[regexp {^[0-9]+\.([0-9]+)} $pos match posnr] == 1} {
		set result $posnr
	}
	return $result
}

proc wcGT {w a b} {
	Debug_FnArgs "wcGT" $w $a $b
	set insert [$w index $a]
	set prompt [$w index $b]
	set insertline [wcGetLineNr $insert]
	set promptline [wcGetLineNr $prompt]
	set insertpos [wcGetPosNr $insert]
	set promptpos [wcGetPosNr $prompt]
	if {($insertline >= $promptline) && ($insertpos > $promptpos)} {
		set result 1
	} else {
		set result 0
	}
	Debug_Msg "wcGT" "insert $insert prompt $prompt result $result"
	return $result
}
proc wcGE {w a b} {
	Debug_FnArgs "wcGT" $w $a $b
	set insert [$w index $a]
	set prompt [$w index $b]
	set insertline [wcGetLineNr $insert]
	set promptline [wcGetLineNr $prompt]
	set insertpos [wcGetPosNr $insert]
	set promptpos [wcGetPosNr $prompt]
	if {($insertline >= $promptline) && ($insertpos >= $promptpos)} {
		set result 1
	} else {
		set result 0
	}
	Debug_Msg "wcGT" "insert $insert prompt $prompt result $result"
	return $result
}

# ============================
#
# Resource file reading
#
# First we try to read from a file given through env var XENVIRONMENT
# If that fails (XENVIRONMENT not set or file unreadable), we try each
# directory in env var XAPPLRESDIR in turn, with as file name either
#  - resourceFileName, if given, or
#  - class, if given, or
#  - app with first letter capitalized
#
# If all fails, we generate an wcError, and then exit.
#
proc wcReadResourceFile {{app main} {class {}} {resourceFileName {}}} {
	global env
	Debug_FnArgs wcReadResourceFile $app $class $resourceFileName
	if {[info exists env(XENVIRONMENT)]} {
		set resourceFile $env(XENVIRONMENT)
		if {[file readable $resourceFile]} {
			if {![catch {option read $resourceFile startup} errMsg]} {
				Debug_Msg wcReadResourceFile $resourceFile
				return
			} else {
				wcError "resource file $resourceFile" $errMsg
				WcExit 1
			}
		}
	}
	if {[info exists env(XAPPLRESDIR)]} {
		set resourceDirs [split $env(XAPPLRESDIR) ':']
		if {$resourceFileName == {}} {
			if {$class != {}} {
				set resourceFileName $class
			} else {
				set resourceFileName [capitalize $app]
			}
		}
		foreach d $resourceDirs {
			set resourceFile [file join $d $resourceFileName]
			if {[file readable $resourceFile]} {
				if {![catch {option read $resourceFile startup} errMsg]} {
					Debug_Msg wcReadResourceFile $resourceFile
					return
				}
			}
		}
	}
	wcError "wcReadResourceFile($resourceFileName)" "- Failed
Problem: No children could be created from Xrm database.
Possible: Resource file not found 
\tperhaps XENVIRONMENT is not set, or doesn't point to the resource file,
\tor XAPPLRESDIR is not set, 
\tor no directory in XAPPLRESDIR contains the resource file 
Possible: top level widget in resource file not named '$resourceFileName'"
	WcExit "." event 1
}
#==================================================================
#
# VPane and HPane building and resizing
#
# Configuration:
# A configuration parameter is wcl(pane,sashwidth):,
# it specifies the space between the pane-children - the background of
# the pane is black, which means that (if wcl(pane,sashwidth) > 0)
# there will be a narrow black line between the pane-children, the width
# of which will be defined by wcl(pane,sashwidth).
# (the name of this parameter is somewhat confusing, unfortunately)
# Useful values: 0 (no space), 1 (narrow line), 2 (wider line)
#
# A configuration parameter is wcl(pane,linewidth):
# it specifies the width of the line that appears (or not, depending on
# the setting of wcl(pane,opaquepaneresize,showline) while resizing
# 'underneath' the sash, and indicates where the pane 'border' will be.
# Useful value: something bigger than 0, best probably 2
#
# A configuration parameter is wcl(pane,linecolor):
# it specifies the color of the line that appears (or not, depending on
# the setting of wcl(pane,opaquepaneresize,showline) while resizing
# 'underneath' the sash, and indicates where the pane 'border' will be.
# Useeful values: black, red
#
# A configuration parameter is wcl(pane,opaquepaneresize):
# it specifies whether or not the panes should already be resized while
# the mouse-button is still pressed (set to 1), or only
# after the mous-button has been released (set to 0).
# Setting to 1 is only advised when the machine/window stuff is 'fast' enough.
# Useful values: 0, 1
#
# A configuration parameter is wcl(pane,opaquepaneresize,showline):
# it specifies whether or not while resizing the line should appear
# 'underneath' the sash, to indicates where the pane 'border' will be.
# Useful values: 0, 1
#
# Probably we should make these resources so they can be specified in
# the resource file, together with the window definition.
# Other things that could be customized via resources:
#  - background color of pane
#  - color of sashes (grips)
#  - size of sashes (grips)
#  - position of sashes (grips) (in percentage of window - currently 0.9)
# 
#
# We create the frame that holds the panes in the wcCreateChild routine;
# in that routine we also create the 'visible' children of the pane.
# However, only after the complete window has been created, we create the
# panes themselves using wcdo_MakePane (called by resp. wcMakeVPane or
# wcMakeHPane): for each child of a pane we create a frame in the pane,
# (this frame we place in the pane, and when the panes have to be resized,
#  we will do this by resizing these auxiliary frames)
# and pack the child into that frame, and create sashes (or grips) on the
# border of each pair of frames.
#
# We only create the auxiliary panes when the complete window has been
# created, because we want to use the size of the children of the pane
# to set the initial size of the auxiliary frames (the idea is that
# wcdo_MakePane does not change the look of the window - it only adds
# auxiliary frames around the pane-children, and adds sashes; even
# if the wcdo_MakePane would never be called, the window is still usable,
# and the only thing that we would loose is that we cannot change the
# relative sizes of the windows in the pane - they remain 'frozen' to
# their initial relative sizes; this also means that (in principle)
# it is sufficient to replace a 'Frame' wcCreate entry by a 'VPane'
# or 'HPane' one to get working panes - the only thing to think of is
# that the part of the child that is to be resized should be the one
# that is pack-ed last into the child (more general: order of
# resizing is reverse of order of pack-ing).
#
# This is why we need the 'update' (not 'update idletasks', that is not
# always (on Suns), or always not (on cygwin/windows), sufficient)
# to be called before we create the auxiliary frames.
# When we place the auxiliary frames in the pane, we use relative sizes
# and relative locations, such that the window will still look good
# when the user resizes the whole window.
#
# When building the sashes (or grips) we attach to them callback
# (event handler) routine wcDoHandleSashMotionEvent that will resize
# the auxiliary frames 'before' and 'after' it, and will invoke the
# event handler routine of the 'next' or 'previuous' sash (or grip)
# when the 'sash that is currently being moved' is moved so far that it
# covers the other sash (this means that one of the other pane-children
# has shrunk to be nonvisible).
#
proc wcHandleSashMotionEventV { W X Y } {
	wcDoHandleSashMotionEvent $W $Y y height
}
proc wcHandleSashMotionEventH { W X Y } {
	wcDoHandleSashMotionEvent $W $X x width
}
proc wcDoHandleSashButtonPressEvent { W } {
	$W configure -relief sunken
}

proc wcDoHandleSashMotionStopEvent { W XY loc size } {
	# puts "wcDoHandleSashMotionStopEvent $W $XY $loc $size"
	global wcl
	$W configure -relief raised
	if {![info exists wcl(pane,info,sash,resized,$W)]} {
		return
	}
	if {! $wcl(pane,info,sash,resized,$W)} {
		return
	}
	set wcl(pane,info,sash,resized,$W) 0
    set root [ split $W . ]
    set nb [ llength $root ]
    incr nb -1
    set root [ lreplace $root $nb $nb ]
    set root [ join $root . ]
    set height [ winfo $size $root ].0
    
    set val [ expr ($XY - [winfo root$loc $root]) /$height ]

    if { $val < 0 } {
    	set val 0
    } elseif { $val > 1.0 } {
    	set val 1.0
    }

    if { $val >= 0 && $val <= 1.0 } {
    	 set oldsashy [expr [winfo $loc $W] + ( [winfo $size $W] / 2.0) ]
         # puts "place $W -rel$loc $val -$loc 0"
         place $W -rel$loc $val -$loc 0
		 place $wcl(pane,info,line,$W) -rel$loc $val -$loc 0
		 lower $wcl(pane,info,line,$W)
         # update ;# just to be sure
         # we do not do the 'just to be sure' update,
         # because, at least on suns, it breaks the `opaque' resize
         update idletasks ;# we _need_ this update to avoid bus errors!!!
         	# and we need it also, to have smoother working!!
    	 set newsashy [expr [winfo $loc $W] + ( [winfo $size $W] / 2.0) ]
    	 if {[winfo exists $wcl(pane,info,prevsash,$W)]} {
		   	 set prevsashy [expr [winfo $loc $wcl(pane,info,prevsash,$W)] + ( [winfo $size $wcl(pane,info,prevsash,$W)] / 2.0) ]
		   	 set prevsashyrel [expr $prevsashy / $height ]
		   	 set prevsashwidth $wcl(pane,sashwidth)
		 } else {
		 	set prevsashy 0
		 	set prevsashyrel 0
		 	set prevsashwidth 0
		 }
    	 if {[winfo exists $wcl(pane,info,nextsash,$W)]} {
		   	 set nextsashy [expr [winfo $loc $wcl(pane,info,nextsash,$W)] + ( [winfo $size $wcl(pane,info,nextsash,$W)] / 2.0) ]
		   	 set nextsashyrel [expr $nextsashy / $height ]
		   	 set nextsashwidth $wcl(pane,sashwidth)
		 } else {
		 	set nextsashy $height
		 	set nextsashyrel 0
		    set nextsashwidth $wcl(pane,sashwidth)
		 }
    	 set oldnexty [winfo $loc $wcl(pane,info,next,$W)]
     	 set newnexty [ expr $newsashy + $wcl(pane,sashwidth) ]
    	 set oldprevy [winfo $loc $wcl(pane,info,prev,$W)]
    	 if { $newsashy <= $prevsashy } {
 	 	 	 # puts "place $wcl(pane,info,prev,$W) -$loc $wcl(pane,sashwidth) -rel$loc $prevsashyrel -$size 0 -rel$size 0"
 	 	 	 place $wcl(pane,info,prev,$W) -$loc $wcl(pane,sashwidth) -rel$loc $prevsashyrel -$size 0 -rel$size 0
	 	 	 set newnextheight [expr $nextsashy - $newsashy]
 	 	 	 set newnextheightrel [expr $newnextheight / $height ]
	   	 	 # puts "place $wcl(pane,info,next,$W) -$loc -$wcl(pane,sashwidth) -rel$loc $val -$size  -$nextsashwidth -rel$size $newnextheightrel"
	   	 	 place $wcl(pane,info,next,$W) -$loc -$wcl(pane,sashwidth) -rel$loc $val -$size  -$nextsashwidth -rel$size $newnextheightrel
 	 	 	 if {[winfo exists $wcl(pane,info,prevsash,$W)]} {
 		  	 	 event generate $wcl(pane,info,prevsash,$W) <B1-ButtonRelease> -root$loc [expr $XY - 10]
 		  	 }
		} elseif { $newsashy >= $nextsashy } {
   	 	 	 set newprevheight [expr $newsashy - $prevsashy ]
   	 	 	 set newprevheightrel [expr $newprevheight / $height ]
   	 	 	 set newprevyrel [expr $prevsashy / $height ]
	   	 	 # puts "place $wcl(pane,info,next,$W) -$loc 0 -rel$loc $val -$size 0 -rel$size 0"
	   	 	 place $wcl(pane,info,next,$W) -$loc 0 -rel$loc $val -$size 0 -rel$size 0
	  	 	 # puts "place $wcl(pane,info,prev,$W) -$loc 0 -rely $newprevyrel -$size -$prevsashwidth -rel$size $newprevheightrel"
	  	 	 place $wcl(pane,info,prev,$W) -$loc 0 -rely $newprevyrel -$size -$prevsashwidth -rel$size $newprevheightrel
 	 	 	 if {[winfo exists $wcl(pane,info,nextsash,$W)]} {
	 	  	 	 event generate $wcl(pane,info,nextsash,$W) <B1-ButtonRelease> -root$loc [expr $XY + 10]
			 }
 	  	 } else {
			 set newprevy $oldprevy
			 set newprevyrel [expr $newprevy / $height ]
 	  	 	 set newprevheight [ expr $newsashy - $prevsashy ]
	  	 	 set newprevheightrel [ expr $newprevheight / $height ]
 	  	 	set newnextheight [ expr $nextsashy - $newsashy ]
	  	 	set newnextheightrel [ expr $newnextheight / $height ]
  	 	 	 # puts "place $wcl(pane,info,prev,$W) -$loc 0 -rel$loc $newprevyrel -$size -$prevsashwidth -rel$size $newprevheightrel"
  	 	 	 place $wcl(pane,info,prev,$W) -$loc 0 -rel$loc $newprevyrel -$size -$prevsashwidth -rel$size $newprevheightrel
	   	 	# puts "place $wcl(pane,info,next,$W) -$loc $nextsashwidth -rel$loc $val -$size -$nextsashwidth -rel$size $newnextheightrel"
	   	 	place $wcl(pane,info,next,$W) -$loc $nextsashwidth -rel$loc $val -$size -$nextsashwidth -rel$size $newnextheightrel
 	  	 }
		foreach sash $wcl(pane,info,sashframes,$W) {
			if { $wcl(pane,info,sash,resized,$sash) } {
	 	  	 	 event generate $sash <B1-ButtonRelease> -root$loc $wcl(pane,info,sash,xy,$sash)
			}
		}
    } else {
    	# puts "wcDoHandleSashMotionStopEvent $val"
    }
}

proc wcDoHandleSashMotionEvent { W XY loc size } {
	# puts "wcDoHandleSashMotionEvent $W $XY $loc $size"
	global wcl
	set wcl(pane,info,sash,resized,$W) 1
	set wcl(pane,info,sash,xy,$W) $XY
    set root [ split $W . ]
    set nb [ llength $root ]
    incr nb -1
    set root [ lreplace $root $nb $nb ]
    set root [ join $root . ]
    set height [ winfo $size $root ].0
    
    set val [ expr ($XY - [winfo root$loc $root]) /$height ]

    if { $val < 0 } {
    	set val 0
    } elseif { $val > 1.0 } {
    	set val 1.0
    }
    if { $val >= 0 && $val <= 1.0 } {
    	 set oldsashy [expr [winfo $loc $W] + ( [winfo $size $W] / 2.0) ]
         # puts "place $W -rel$loc $val -$loc 0"
         place $W -rel$loc $val -$loc 0
		 place $wcl(pane,info,line,$W) -rel$loc $val -$loc 0
		 if { (! $wcl(pane,opaquepaneresize)) ||
		      $wcl(pane,opaquepaneresize,showline) } {
		 	raise $wcl(pane,info,line,$W)
		 }
		 raise $W ;# make sure that the sash that is being dragged is on top!
         # update ;# just to be sure
         # we do not do the 'just to be sure' update,
         # because, at least on suns, it breaks the `opaque' resize
         update idletasks ;# we _need_ this update to avoid bus errors!!!
         	# and we need it also, to have smoother working!!
    	 set newsashy [expr [winfo $loc $W] + ( [winfo $size $W] / 2.0) ]
    	 if {[winfo exists $wcl(pane,info,prevsash,$W)]} {
		   	 set prevsashy [expr [winfo $loc $wcl(pane,info,prevsash,$W)] + ( [winfo $size $wcl(pane,info,prevsash,$W)] / 2.0) ]
		   	 set prevsashyrel [expr $prevsashy / $height ]
		   	 set prevsashwidth $wcl(pane,sashwidth)
		 } else {
		 	set prevsashy 0
		 	set prevsashyrel 0
		 	set prevsashwidth 0
		 }
    	 if {[winfo exists $wcl(pane,info,nextsash,$W)]} {
		   	 set nextsashy [expr [winfo $loc $wcl(pane,info,nextsash,$W)] + ( [winfo $size $wcl(pane,info,nextsash,$W)] / 2.0) ]
		   	 set nextsashyrel [expr $nextsashy / $height ]
		   	 set nextsashwidth $wcl(pane,sashwidth)
		 } else {
		 	set nextsashy $height
		 	set nextsashyrel 0
		    set nextsashwidth $wcl(pane,sashwidth)
		 }
    	 set oldnexty [winfo $loc $wcl(pane,info,next,$W)]
     	 set newnexty [ expr $newsashy + $wcl(pane,sashwidth) ]
    	 set oldprevy [winfo $loc $wcl(pane,info,prev,$W)]
    	 if { $newsashy < $prevsashy } {
 	 	 	 # puts "place $wcl(pane,info,prev,$W) -$loc $wcl(pane,sashwidth) -rel$loc $prevsashyrel -$size 0 -rel$size 0"
 	 	 	 if {$wcl(pane,opaquepaneresize)} {
 	 	 	 	place $wcl(pane,info,prev,$W) -$loc $wcl(pane,sashwidth) -rel$loc $prevsashyrel -$size 0 -rel$size 0
 	 	 	 }
	 	 	 set newnextheight [expr $nextsashy - $newsashy]
 	 	 	 set newnextheightrel [expr $newnextheight / $height ]
	   	 	 # puts "place $wcl(pane,info,next,$W) -$loc -$wcl(pane,sashwidth) -rel$loc $val -$size  -$nextsashwidth -rel$size $newnextheightrel"
 	 	 	 if {$wcl(pane,opaquepaneresize)} {
	   	 	 	place $wcl(pane,info,next,$W) -$loc -$wcl(pane,sashwidth) -rel$loc $val -$size  -$nextsashwidth -rel$size $newnextheightrel
	   	 	 }
 	 	 	 if {[winfo exists $wcl(pane,info,prevsash,$W)]} {
 		  	 	 event generate $wcl(pane,info,prevsash,$W) <B1-Motion> -root$loc $XY
 		  	 }
		} elseif { $newsashy > $nextsashy } {
   	 	 	 set newprevheight [expr $newsashy - $prevsashy ]
   	 	 	 set newprevheightrel [expr $newprevheight / $height ]
   	 	 	 set newprevyrel [expr $prevsashy / $height ]
	   	 	 # puts "place $wcl(pane,info,next,$W) -$loc 0 -rel$loc $val -$size 0 -rel$size 0"
 	 	 	 if {$wcl(pane,opaquepaneresize)} {
	   	 	 	place $wcl(pane,info,next,$W) -$loc 0 -rel$loc $val -$size 0 -rel$size 0
	   	 	 }
	  	 	 # puts "place $wcl(pane,info,prev,$W) -$loc 0 -rely $newprevyrel -$size -$prevsashwidth -rel$size $newprevheightrel"
 	 	 	 if {$wcl(pane,opaquepaneresize)} {
	  	 	 	place $wcl(pane,info,prev,$W) -$loc 0 -rely $newprevyrel -$size -$prevsashwidth -rel$size $newprevheightrel
	  	 	 }
 	 	 	 if {[winfo exists $wcl(pane,info,nextsash,$W)]} {
	 	  	 	 event generate $wcl(pane,info,nextsash,$W) <B1-Motion> -root$loc $XY
			 }
 	  	 } else {
			 set newprevy $oldprevy
			 set newprevyrel [expr $newprevy / $height ]
 	  	 	 set newprevheight [ expr $newsashy - $prevsashy ]
	  	 	 set newprevheightrel [ expr $newprevheight / $height ]
 	  	 	set newnextheight [ expr $nextsashy - $newsashy ]
	  	 	set newnextheightrel [ expr $newnextheight / $height ]
  	 	 	 # puts "place $wcl(pane,info,prev,$W) -$loc 0 -rel$loc $newprevyrel -$size -$prevsashwidth -rel$size $newprevheightrel"
 	 	 	 if {$wcl(pane,opaquepaneresize)} {
  	 	 	 	place $wcl(pane,info,prev,$W) -$loc 0 -rel$loc $newprevyrel -$size -$prevsashwidth -rel$size $newprevheightrel
  	 	 	 }
	   	 	# puts "place $wcl(pane,info,next,$W) -$loc $nextsashwidth -rel$loc $val -$size -$nextsashwidth -rel$size $newnextheightrel"
 	 	 	 if {$wcl(pane,opaquepaneresize)} {
	   	 		place $wcl(pane,info,next,$W) -$loc $nextsashwidth -rel$loc $val -$size -$nextsashwidth -rel$size $newnextheightrel
	   	 	}
 	  	 }
    }
}

proc wcMakeVPane { w childpath } {
	wcdo_MakePane $w $childpath y height x width e %Y
}
proc wcMakeHPane { w childpath } {
	wcdo_MakePane $w $childpath x width y height s %X
}
proc wcdo_MakePane { w childpath loc size otherloc othersize sanchor spos } {
	global wcl
	set pane(children) [winfo children $childpath]
	if {[llength $pane(children)] > 1} {
		set pane(size) 0
		set pane(i) 0
		set pane(cumsize) 0
		foreach panechild $pane(children) {
			set pane(size,$panechild) [winfo $size $panechild]
			incr pane(cumsize) $pane(size,$panechild)
			# puts "size $panechild: $pane(size,$panechild)"
		}
		set pane(size) [winfo $size $w]
		# puts "=="
		# puts "size $w: $pane(size); cumsize: $pane(cumsize)"
		# puts ""
		if {$pane(size) != $pane(cumsize)} {
			# wcWarning "in wcdo_MakePane: sum(sizeof(pane-children)) (=$pane(cumsize)) != sizeof(pane) (=$pane(size))"
			incr pane(size,[lindex $pane(children) 0]) \
				[expr $pane(size) - $pane(cumsize)]
		}
		set pane(cumsize) 0
		set pane(prev) [lindex $pane(children) 0]
		set pane(prevframe) [frame $childpath._genf$pane(i)]
		place $pane(prevframe) -x 0 -y 0 \
			-rel$loc 0 -rel$otherloc 0 \
			-rel$size [expr $pane(size,$pane(prev)) / $pane(size).0] \
			-$size -$wcl(pane,sashwidth) -$othersize 0 -rel$othersize 1 \
			-anchor nw ;#-bordermode ignore
		pack $pane(prev) -in $pane(prevframe) -anchor center -expand 1 \
			-fill both -side top
		raise $pane(prev)
		incr pane(cumsize) $pane(size,$pane(prev))
		# puts "cumsize $pane(cumsize)"
		set pane(prevsash) 0
		set sashframes {}
		foreach panechild [lrange $pane(children) 1 end] {
			incr pane(i)
			set pane(next) $panechild
			set pane(nextframe) [frame $childpath._genf$pane(i)]
			set pane(sashframe) \
				[frame $childpath._gens$pane(i) -borderwidth 2 -relief raised]
			set pane(lineframe) \
				[frame $childpath._genl$pane(i) -borderwidth 2 -relief raised \
				 -background $wcl(pane,linecolor)  ]
			set pane(locrel) [expr $pane(cumsize) / $pane(size).0]
			set pane(sizerel) [expr $pane(size,$pane(next)) / $pane(size).0]
			place $pane(nextframe) -x 0 -y 0 \
				-rel$loc $pane(locrel) -rel$otherloc 0 \
				-rel$size $pane(sizerel) -$size -$wcl(pane,sashwidth) \
				-$othersize 0 -rel$othersize 1 -anchor nw ;# -bordermode ignore
			pack $pane(next) -in $pane(nextframe) -anchor center -expand 1 \
				-fill both -side top
			raise $pane(next) ;# or else the frame obscures the 'contents'
			lower $pane(lineframe)
			raise $pane(sashframe)
			incr pane(cumsize) $pane(size,$pane(next))
			# puts "cumsize $pane(cumsize)"
			place $pane(sashframe) -$otherloc 0 -rel$otherloc 0.9 \
				-$loc 0 -rel$loc $pane(locrel) \
				-width 10 -height 10 -anchor $sanchor ;#-bordermode ignore 
			place $pane(lineframe) -$otherloc 0 -rel$otherloc 1 \
				-$loc 0 -rel$loc $pane(locrel) \
				-rel$othersize 1 -$othersize 0 -$size $wcl(pane,linewidth) -anchor $sanchor ;#-bordermode ignore 
			
			lappend sashframes $pane(sashframe)
			bind $pane(sashframe) <ButtonPress-1> \
				[list wcDoHandleSashButtonPressEvent %W]
			bind $pane(sashframe) <B1-Motion> \
				[list wcDoHandleSashMotionEvent %W $spos $loc $size]
			bind $pane(sashframe) <B1-ButtonRelease> \
				[list wcDoHandleSashMotionStopEvent %W $spos $loc $size]
			bind $pane(sashframe) <Enter> \
				[list %W configure -background $wcl(pane,sashhighlightbackground)]
			bind $pane(sashframe) <Leave> \
				[list %W configure -background $wcl(pane,sashbackground)]
			set wcl(pane,info,line,$pane(sashframe)) $pane(lineframe)
			set wcl(pane,info,prevsash,$pane(sashframe)) $pane(prevsash)
			set wcl(pane,info,nextsash,$pane(prevsash)) $pane(sashframe)
			set wcl(pane,info,prev,$pane(sashframe)) $pane(prevframe)
			set wcl(pane,info,next,$pane(sashframe)) $pane(nextframe)
			set pane(prevframe) $pane(nextframe)
			set pane(prevsash) $pane(sashframe)
		}
		set wcl(pane,info,nextsash,$pane(prevsash)) 0
		foreach sash $sashframes {
			set wcl(pane,info,sashframes,$sash) $sashframes
			set wcl(pane,info,sash,resized,$sash) 0
		}
		# puts "pack append $w $childpath top expand fill"
		pack append $w $childpath {top expand fill}
	}
}
