#! @PERL@ --
#
# $Id: allinst.pl,v 1.1.2.4 2007/05/24 14:14:20 belinfan Exp $
#
# usage: allinst [ -s seed ] [ -f configFile ]
#
# variant of the instantiator that returns all possible instantiations
# instead of only a single one.
#
# it reads the messages from stdin, and instatiates
# variables in these messages.
# Which variables have to be instantiated by values from which domains can be expressed
# in a separate configuration file using the (-f configFile) flag.
# The instatiation values are chosen randomly from the domains given in the config file;
# the seed of the random number generator can be set using the (-s seed) flag.
# 

# What do we want to do?
# We initialise:
#  - We read the config file
#     - Format:
#           type : position : prefix : regexp : domain
#       - where leading and trailing whitespace is removed in the lines
#       - where continuation lines (ending in \ ) are allowed
#         Note: the \ is replaced by a space when joining lines
#       - where comment lines are ignored
#         (lines starting with optional whitespace followed by #)
#       - where empty lines are ignored
#         (really empty lines and lines containing only whitespace)
#       - where leading and trailing whitespace in each of the fields
#         (type, position, prefix, regexp, domain) is ignored!!!
#       - where position may be * to indicate any position
#       - domain is either something like {val1,val2,val3,val4}
#         i.e. a comma-separated list of values, enclosed in { and }
#         where the values themselves do not contain comma's,
#         where there should only one comma (and nothing else)
#         between each pair of values, and there should be nothing
#         between { and first value nor between last value and }
#       - or domain is something like ( e1(e2,e3)  b1 c1 d1(e3,f3(g,h)) )
#         i.e. a whitespace-separated list of values, enclosed in ( and )
#         where the values themselves do not contain whitespace,
#         where the amount of whitespace between each pair of values
#         does not matter, nor whitespace directly after ( and/or directly
#         before )
#       - the prefix can be a regexp;
#          IMPORTANT: all lines belonging to the same instantiation group
#          should have _the_same_ prefix (because, as soon as we can make a
#          instantiation using one of the prefixes, the rules of the other
#          prefixes will not be tried)
#       - where we lines with wrong number of fields (and give a warning)
#       - where we warn for empty fields
#  - We initialise the random number generator
# We go to our main loop:
#  - We read lines from standard input.
#     - If the start of a line matches one of the prefixes,
#       then we split the line into fields on TAB, and look only at event field,
#       which we scan for variables;
#       - with each varable we associate 'context' information:
#          - the prefix (part?) of the command that does not contain variables
#            One way to do this is the following:
#            let the config info consist of tuples < vartype, varpos, prefix, regexp, domain >
#            and simply go for each of the prefixes through its list of regexps for the
#            type and position of the variable until a regexp matches with the (preprocessed?)
#            line. As soon as we have a match for a certain prefix, we will not look at the
#            rules/regexps for the other prefixes.
#          - the position of the variable in the command
#          - the type of the variable
#       - we lookup the context in the config information, which should give us a domain
#         from which we choose a value (here we also take wildcards (*) into account)
#       - we associate the chosen value with the variable
#       - if we have done this for all variables, we output the line (with vars substituted)
#       - we check if we have made substitutions, and if so, we check if the line still
#         contains variables: if so, we loop, trying to find values for those variables.
#       - the 'mtype' type is special: if we have found a value for it, we don't try to
#         work on the other variables in the line, but immediately substitute the value,
#         and loop, so the value found for the mtype variable can be used as context
#         for the other variables.
#       - we instantiate also variables that are 'embedded' in a larger expression
#     - If a line does not match a prefix, we echo it on stdout
#
# Example config file (without # at start of line)
#
#		# ==================================
#		# empty lines and comment lines are ignored.
#		# comment lines are lines that start with
#		# (optional whitespace followed by) a hash sign (#).
#		byte:2:C_INPUT:from_upper!JOIN!.*!.*:{1,2,3,4}
#		byte:3:C_INPUT:from_upper!JOIN!.*!.*:{1,2}
#		
#		# NOTE: _both_ lines of this regexp group have same prefix;
#		# if one line has C_INPUT and the other C_INPU(S?) then
#		# only one instantiation will be done even if an input line
#		# contains two var_byte fields...
#		byte:2:C_INPUT(S?):from_upper!LEAVE!.*!.*:{1,2,3,4}
#		byte:3:C_INPUT(S?):from_upper!LEAVE!.*!.*:{1,2}
#		
#		byte:2:C_INPUT:from_upper!DREQ!.*!.*:{1,2,3,4}
#		byte:3:C_INPUT:from_upper!DREQ!.*!.*:{1}
#		
#		byte:2:C_INPUT:from_lower!PDU_JOIN!.*!.*!.*!.*:{1,2,3,4}
#		byte:3:C_INPUT:from_lower!PDU_JOIN!.*!.*!.*!.*:{1,2}
#		byte:4:C_INPUT:from_lower!PDU_JOIN!.*!.*!.*!.*:{0,2}
#		byte:5:C_INPUT:from_lower!PDU_JOIN!.*!.*!.*!.*:{1}
#		# ==================================
#
# ======================================

use strict;

require 5.002;
use Getopt::Long;

# ======================================

my (
	$prog,
	%optcl, $result, $debug, $configFile, $seed,
	%prefixes, %regexps, %domains,
   );

# ======================================

$prog = 'instantiator';
$| = 1;
$seed = 0;
$configFile = 'instconfig.txt';

# ======================================

%optcl = ();
$result = GetOptions(
			\%optcl,
			"f=s",
			"s=i",
			"debug:i",
		);

$debug = 1 if ((defined $optcl{"debug"}) && ! $optcl{"debug"});
$debug = $optcl{"debug"} if ((defined $optcl{"debug"}) && $optcl{"debug"});
$configFile = $optcl{"f"} if (defined $optcl{"f"});
$seed = $optcl{"s"} if (defined $optcl{"s"});

# ======================================

init_random($seed);
read_configuration($configFile);
main_loop();

# ======================================

sub init_random {
	my($seed) = @_;
	srand($seed);
}

# ======================================

sub read_configuration {
	my($file) = @_;
	my(@fields,$fieldslen,$linenr,$buffered,$type,$pos,$prefix,$regexp,$domain,$rk,$dk,@regexparr,$k,$r,$d);
	open(H, $file)
		|| die "$prog: Could not read configuration file $file: $!\n";
	$linenr = 0;
	$buffered = '';
	while(<H>) {
		$linenr++;
		chomp;				# remove trailing newline junk
		$_ = stripleadtrailws($_);
		next if /^\s*$/;	# skip empty lines
		next if /^\s*#/;	# skip comment lines
		if (/\\$/) {		# continuation line
			print STDERR " - continuation line $linenr\n"
				if $debug;
			s/\\$/ /;		# NOTE: this inserts seperating whitespace
			$buffered .= $_;
			next;
		} else {
			print STDERR " - no continuation line $linenr\n"
				if $debug;
			$_ = $buffered . $_;# NOTE: no need to insert seperating whitespace (already done)
			$buffered = '';
		}
		@fields = split(':');
		$fieldslen = @fields;
		if ($fieldslen != 5) {
			warn "$prog: skipping config line $file:$linenr : has $fieldslen fields (instead of 5)\n";
			next;
		}
		$type = stripleadtrailws($fields[0]);
		$pos = stripleadtrailws($fields[1]);
		$prefix = stripleadtrailws($fields[2]);
		$regexp = stripleadtrailws($fields[3]);
		$domain = stripleadtrailws($fields[4]);
		warn "$prog: empty 'type' field at config line $file:$linenr\n"
			if $type =~ /^\s*$/;
		warn "$prog: empty 'position' field at config line $file:$linenr\n"
			if $pos =~ /^\s*$/;
		warn "$prog: empty 'prefix' field at config line $file:$linenr\n"
			if $prefix =~ /^\s*$/;
		warn "$prog: empty 'regexp' field at config line $file:$linenr\n"
			if $regexp =~ /^\s*$/;
		warn "$prog: empty 'domain' field at config line $file:$linenr\n"
			if $domain =~ /^\s*$/;
		$prefixes{$prefix} = 1;
		$rk = "$type$;$pos$;$prefix";
		@regexparr = split($;, $regexps{$rk});
		push(@regexparr, $regexp);
		$regexps{$rk} = join($;, @regexparr);
		$dk = "$type$;$pos$;$prefix$;$regexp";
		$domains{$dk} = $domain;
	}
	close(H);
	if ($debug) {
		foreach $k (keys %regexps) {
			print STDERR "$k :\n";
			foreach $r (split($;,$regexps{$k})) {
				print STDERR "\t$r\n";
			}
		}
		foreach $k (keys %domains) {
			print STDERR "$k :\n";
			foreach $d (split($;,$domains{$k})) {
				print STDERR "\t$d\n";
			}
		}
	}
}

# ======================================

sub main_loop {
	my($ch, @in, @out, @tmp, @newresultlines, $l, @result, $r, $pfx, $changed);
	while(<>) {
		chomp($_);
		print STDERR "instant: $_\n";
		@in = ($_);
		@tmp = ();
		@out = ();
		while (@in > 0) {
			foreach $l (@in) {
				print STDERR "    --- in2 = ($l)\n"
					if $debug;
				($ch, @newresultlines) = handle_line($l);
				print STDERR "    --- ch $ch newresultline = (@newresultlines)\n"
					if $debug;
				$changed = $changed || $ch;
				if ($ch) {
					@tmp = (@tmp, @newresultlines);
				} else {
					@out = (@out, @newresultlines);
				}
			}
			@in = @tmp;
			@tmp = ();
		}
		if (@out > 0) {
			$pfx = $_;
			$pfx =~ s/^\s*([^\s]+).*$/\1/;
			$pfx =~ s/^C_/A_/;
			print $pfx.'_START'."\n";
			foreach $r (@out) {
				$r =~ s/^C_/A_/;
				print $r."\n";
			}
			print $pfx.'_END'."\n";
		} else {
			print $_."\n";
		}
	}
}
		
sub handle_line {
	my($line,$prefix,$lineprefix,@items,$item,@resultitems,@event,$event,$eventarg,$eventprefix,$resultline,$madesubst,$mtype_substituted,$changed, $n, $ni, @tresultitems, @resultlines, @newitem, $re);
	
	my($line) = @_;
	
		print STDERR "    --- line = ($line)\n"
			if $debug;
		foreach $prefix (keys %prefixes) {
			print STDERR "\t--- trying ($prefix)\n"
				if $debug;
			if (/(^$prefix\s+)/i) {
				$lineprefix = $1;
				chomp;
				$line =~ s/^$lineprefix//i;
				$changed = 0;
				@items = split('\t', $line);
				@resultitems = ();
				foreach $item (@items) {
					print STDERR "\t--- item ($item)\n"
						if $debug;
					@newitem = ();
					if ($item =~ /^(event=)(.*)$/) {
						($eventprefix, $event) = ($1, $2);
						print STDERR "\t--- item ($eventprefix)($event)\n"
							if $debug;
						$mtype_substituted = 1;
						# while ($mtype_substituted) {
							$mtype_substituted = 0;
							$eventarg = $event;
							$eventarg =~ s/\s//g;
							($madesubst, $mtype_substituted, @event) =
								handle_event($prefix, $eventarg, $mtype_substituted);
							if ($madesubst && ($_=~ /var(_(\w+))/)) {
								$mtype_substituted = 1;
							}
							if ($madesubst) {
								$changed = 1;
							}
						# }
						if (1 || $changed) {
							foreach $n (@event) {
								$ni = $eventprefix.$n;
								push(@newitem, $ni);
								print STDERR "\t--- item result ($ni)\n"
									if $debug;
							}
						}
					} elsif ($item =~ /^(predicates=)(.*)$/) {
						@newitem = ('predicates=');
					}
					if (@newitem == 0) {
						@newitem = ($item);
					}
					@tresultitems = @resultitems;
					@resultitems = ();
					if (@tresultitems != 0) {
						foreach $re (@tresultitems) {
							foreach $ni (@newitem) {
								push(@resultitems, $re."\t".$ni);
							}
						}
					} else {
						foreach $ni (@newitem) {
							push(@resultitems, $ni);
						}
					}
				}
				if (1|| $changed) {
					foreach $line (@resultitems) {
						$resultline = $lineprefix.$line;
						push(@resultlines, $resultline);
					}
				}
			}
		}
	return ($changed, @resultlines);
}

# ======================================
sub handle_event {
	my($prefix,$line) = @_;
	my(@line,@tline,@part,$pos,$type,$p,$rk,$rk1,$rk2,@regexps,@regexps1,@regexps2,$regexp,$dk,$dk1,$dk2,$val,$result,$resultline,$madesubst,$mtype_substituted, @val, @v, @vv, $v, $vv, @res, $l, $r);
	
	print STDERR "\t    --- working on ($prefix)($line)\n"
		if $debug;
	@part = split('!', $line);
	$pos = -1;
	undef @line;
	$madesubst = 0;
	@res = ();
	foreach $p (@part) {
		$pos++;
		$result = $p;
		print STDERR "\t\t--- looking at $pos ($p)($line)\n"
			if $debug;
		if ($p =~ /^var(_(\w+))?$/) {
			$type = $2;
			print STDERR "\t\t\t--- trying match $pos $type ($line)\n"
				if $debug;
			$rk1 = "$type$;$pos$;$prefix";
			$rk2 = "$type$;*$;$prefix";
			@regexps1 = split($;,$regexps{$rk1});
			@regexps2 = split($;,$regexps{$rk2});
			@regexps = ( @regexps1, @regexps2);
			foreach $regexp (@regexps) {
				@res = ();
				$dk1 = "$type$;$pos$;$prefix$;$regexp";
				$dk2 = "$type$;*$;$prefix$;$regexp";
				# $regexp =~ s/\./[^!]/g;
				print STDERR "\t\t\t\t--- trying regexp ($regexp)\n"
					if $debug;
				if ($line =~ /^$regexp$/i) {
					print STDERR "\t\t\t\t\t--- found match ($regexp)\n"
						if $debug;
					@val = get_values($type, $domains{$dk1});
					foreach $v (@val) {
						if ($v =~ /^var_$type$/) {
							@v = get_values($type, $domains{$dk2});
						} else {
							@v = ($v);
						}
						@res = (@res, @v);
						foreach $vv (@v) {
							print STDERR "\t\t\t\t\t--- chosen vv value ($v)\n"
								if $debug;
						}
					}
					$madesubst = 1;
					if ($type eq 'mtype') {
						$mtype_substituted = 1;
						print STDERR "\t\t\t\t\t--- substituted mtype ($type)\n"
						if $debug;
					}
					last;
				}
			}
		} elsif ($p =~ /var(_([a-zA-Z0-9_]+))?/) {
			$type = $2;
			print STDERR "\t\t\t--- trying match $pos $type ($line)\n"
				if $debug;
			$rk1 = "$type$;$pos$;$prefix";
			$rk2 = "$type$;*$;$prefix";
			@regexps1 = split($;,$regexps{$rk1});
			@regexps2 = split($;,$regexps{$rk2});
			@regexps = ( @regexps1, @regexps2);
			foreach $regexp (@regexps) {
				@res = ();
				$dk1 = "$type$;$pos$;$prefix$;$regexp";
				$dk2 = "$type$;*$;$prefix$;$regexp";
				# $regexp =~ s/\./[^!]/g;
				print STDERR "\t\t\t\t--- trying regexp ($regexp)\n"
					if $debug;
				if ($line =~ /^$regexp$/i) {
					print STDERR "\t\t\t\t\t--- found match ($regexp)\n"
						if $debug;
					@val = get_values($type, $domains{$dk1});
					print STDERR "\t\t\t\t\t--- found values (@val )  ".choose_value($type, $domains{$dk1})."\n"
						if $debug;
					foreach $v (@val) {
						if ($v =~ /^var_$type$/) {
							print STDERR "\t\t\t\t\t--- chosen vv v value ($v)\n"
								if $debug;
							@vv = get_values($type, $domains{$dk2});
						} else {
							@vv = ($v);
						}
						foreach $vv (@vv) {
							print STDERR "\t\t\t\t\t--- chosen vv value ($vv)\n"
								if $debug;
							$result = $p;
							$result =~ s/var(_([a-zA-Z0-9_]+(\$[0-9]+)?))?/$vv/;
							print STDERR "\t\t\t\t\t--- substituted value ($result)\n"
								if $debug;
							$madesubst = 1;
							$mtype_substituted = 1;
							print STDERR "\t\t\t\t\t--- substituted mtype ($type)\n"
								if $debug;
							@res = (@res, $result);
						}
					}
					last;
				}
			}
		} else {
			@res = ($p);
		}
		print STDERR "\t\t--- res (@res)\n"
			if $debug;
		if (@res == 0) {
			@res = ($result);
		}
		@tline = @line;
		@line = ();
		if (@tline != 0) {
			foreach $l (@tline) {
				foreach $r (@res) {
					push(@line, $l.'!'.$r);
				}
			}
		} else {
			foreach $r (@res) {
				push(@line, $r);
			}
		}
	}
	foreach $l (@line) {
		print STDERR "\t--- hndle ev result ($l)\n"
			if $debug;
	}
	return ($madesubst, $mtype_substituted, @line);
}

# ======================================

sub choose_value {
	my($type, $domain) = @_;
	my($result,$values,@values,$nrof,$index) = "var_$type";
	if ($domain =~ /^{(.*)}$/) {
		$values = $1;
		@values = split(',',$values);
		$nrof = @values;
		$index = int(rand($nrof));
		$result = $values[$index];
	} elsif ($domain =~ /^\((.*)\)$/) {
		$values = stripleadtrailws($1);
		@values = split(' ',$values);
		$nrof = @values;
		$index = int(rand($nrof));
		$result = $values[$index];
	}
	return $result;
}

sub get_values {
	my($type, $domain) = @_;
	my($result,$values,@values,$nrof,$index) = "var_$type";
	if ($domain =~ /^{(.*)}$/) {
		$values = $1;
		@values = split(',',$values);
	} elsif ($domain =~ /^\((.*)\)$/) {
		$values = stripleadtrailws($1);
		@values = split(' ',$values);
	} else {
		@values = ($result);
	}
	return @values;
}

# ======================================

sub stripleadtrailws {
	my($l) = @_;
	$l =~ s/^\s*//g;
	$l =~ s/\s*$//g;
	return $l;
}
# ======================================
