# =========================================================================
# $Id: unify.tcl,v 1.1.2.1 2000/08/28 12:12:37 belinfan Exp $
# =========================================================================


# The unify routine takes an open term and a (partially?)
# closed one, and an optional list of free variables,
#  and reports
#  - if they can be unified (i.e. if they are structurally
#    equal, modulo subterms that instantiate variables)
#  - if so, the list of bindings that makes up the
#    unification, i.e. the bindings that transform the
#    open term into the (partially) closed one.
# So, the result will be either:
#   the singleton 0, indicating false, or
#   the tuple { 1 {list-of-bindings} }
#
# The idea is that both input terms are in the event/term
# syntax, as expected.
# Furthermore, currently we can syntactically
# identify variables in the open term because
# they have one of two forms:
#  - var_<type>       where <type> is a type name
#  - Smile_<nr>_<nr>  where <nr> is an integer

proc unify {openterm closedterm {free_vars {}} } {
	# first, check the gate names and the
	# number of experiment options, to rule
	# out the simples failures.
	if {[llength $openterm] !=
	    [llength $closedterm]} {
		return [list 0]
	}
	set opengate [lindex [lindex $openterm 0] 0]
	set closedgate [lindex [lindex $closedterm 0] 0]
	if {[string compare [string tolower $opengate] [string tolower $closedgate]] != 0} {
		return [list 0]
	}
	
	# now, we basically want to check structural
	# equivalence modulo variables/instantiation,
	# and build up a list of bindings, if necessary.
	set bindings {}
	set cur_exp 1 ;# we already checked the gates
	set past_last_experiment [llength $openterm]
	while {$cur_exp < $past_last_experiment} {
		set tmp [unify_experiments [lindex $openterm $cur_exp] [lindex $closedterm $cur_exp] $free_vars]
		if {![lindex $tmp 0]} {
			return [list 0]
		} else {
			foreach b [lindex $tmp 1] {
				lappend bindings $b
			}
		}
		incr cur_exp
	}
	return [list 1 $bindings]
}

# ==========================================

proc unify_experiments {open_exp closed_exp free_vars} {
	# debugMsg "unify_experiments ($open_exp) ($closed_exp)..."
	set open_exp_name [lindex $open_exp 0]
	set open_exp_sub [lindex $open_exp 1]
	set closed_exp_name [lindex $closed_exp 0]
	set closed_exp_sub [lindex $closed_exp 1]

	# see if we can find a variable substitution
	# NOTE: if the closed_term also looks like a variable,
	# we DO NOT make a substitution.
	# Maybe, we should relax this a bit, and only NOT make
	# a substitution if the closed_term varaible-name is
	# identical to the open_term variable-name...
	if {([is_variable_name $open_exp_name $free_vars]) &&
	    ([llength $open_exp_sub] == 0) &&
	    ! (([is_variable_name $closed_exp_name $free_vars]) &&
	       ([llength $closed_exp_sub] == 0))} {
	    set result [list 1 [list [list $open_exp_name $closed_exp]]]
		# debugMsg "unify_experiments ($open_exp) ($closed_exp): $result"
		return $result
	}

	# not a variable subsitution,
	# see if we can find a match on this level,
	# and then go on recursively:
	#  - check operator name
	if {[string compare [string tolower $open_exp_name] [string tolower $closed_exp_name]] != 0} {
	    set result [list 0]
		# debugMsg "unify_experiments ($open_exp) ($closed_exp): $result"
		return $result
	}
	#  - check respective number of subterms
	if {[llength $open_exp_sub] !=
	    [llength $closed_exp_sub]} {
	    set result [list 0]
		# debugMsg "unify_experiments ($open_exp) ($closed_exp): $result"
		return $result
	}
	#  - recursively check the subterms
	# now, we basically want to check structural
	# equivalence modulo variables/instantiation,
	# and build up a list of bindings, if necessary.
	set bindings {}
	set cur_exp 0
	set past_last_experiment [llength $open_exp_sub]
	while {$cur_exp < $past_last_experiment} {
		set tmp [unify_experiments [lindex $open_exp_sub $cur_exp] [lindex $closed_exp_sub $cur_exp]  $free_vars]
		if {![lindex $tmp 0]} {
		    set result [list 0]
			# debugMsg "unify_experiments ($open_exp) ($closed_exp): $result"
			return $result
		} else {
			foreach b [lindex $tmp 1] {
				lappend bindings $b
			}
		}
		incr cur_exp
	}
	set result [list 1 $bindings]
	# debugMsg "unify_experiments ($open_exp) ($closed_exp): $result"
	return $result
}

# ==================================

proc is_variable_name {var free_vars} {
	if {[regexp -nocase -- "smile_\[0-9]+_\[0-9]+" $var]} {
		return 1
	} elseif {[regexp -nocase -- "var_\[a-zA-Z0-9_]+" $var]} {
		return 1
	} else {
		return 0
	}
}

# ==================================

