#! @EXPECT@ --
#! /Utils/bin/tclsh
#! /usr/local/bin/expect --

# NOTE: this script only needs tcl(sh); it does not use expect
# features. The only reason that we use the expect interpreter
# here is that the configure script of TorX finds the location of
# expect already for us, and (AFAIK) does not (yet?) look for tclsh.

proc find_user {host} {
	global env

	if {[info exists env(USER)]} {
		set user $env(USER)
	} elseif {[info exists env(LOGNAME)]} {
		set user $env(LOGNAME)
	} else {
		set user unknown
	}
	return $user
}

proc find_display {host} {
	global env

	if {[info exists env(DISPLAY)]} {
		set display $env(DISPLAY)
	} else {
		set display ":0"
	}
	regsub {^:} $display "$host:" display
	regsub {^unix:} $display "$host:" display
	regsub {\.[0-9]*$} $display "" display
	return $display
}

proc mk_filename {dir pfx user display} {
	set fname [file join $dir "$pfx-$user-$display"]
	regsub -all : $fname _ fname
	return $fname
}

proc try_connect { fname do_warn } {
	if {![catch {open $fname r} fid]} {
		if {[gets $fid line] > 0} {
			set words [split $line]
			set ipNr [lindex $words 0]
			set hostName [lindex $words 1]
			set portNr [lindex $words 2]
			if {![catch {socket $ipNr $portNr} socket]} {
				return $socket
			} else {
				warning "could not connect tcp!$ipNr!$portNr: $socket"
			}
		}
		close $fid
	} elseif {$do_warn} {
		warning "$fid"
	}
	return 0
}

# exp_internal 1
set debug 0

set prog aniwait

# we use output to make sure that everything is flushed!
# we do this to avoid deadlocks due to buffers being full...
proc output {chan msg} {
	puts $chan $msg
	flush $chan
}

proc fatal {s {code {0}}} {
	global prog
	set s [string trim $s]
	output stderr "$prog: $s"
	cleanup_exit $code
}
proc message {s} {
	global prog
	set s [string trim $s]
	output stdout "$prog: $s"
}
proc warning {s} {
	global prog
	set s [string trim $s]
	output stderr "$prog: $s"
}

proc debug {s} {
	global prog
	global debug
	if {$debug} {
		set s [string trim $s]
		output stderr "$prog: $s"
	}
}

proc cleanup_exit {{code {0}}} {
	global state
	if {[info exists state(stderr)]} {
		eval fconfigure stderr $state(stderr)
	}
	if {[info exists state(stdout)]} {
		eval fconfigure stdout $state(stdout)
	}
	if {[info exists state(stdin)]} {
		eval fconfigure stdin $state(stdin)
	}
	exit $code
}

proc configure_fids {socket} {
	fconfigure $socket -buffering line  -translation {binary binary}
	fconfigure stdin -buffering line \
		                   -translation {binary binary}
}

proc filevent_fids {socket remoteHost remotePort waitRemoteClose} {
	fileevent $socket readable \
	    	[list getRemote $remoteHost $remotePort $socket]
	fileevent stdin readable \
	 	[list sendRemote $remoteHost $remotePort $socket $waitRemoteClose]
	vwait forever
}

proc setUp {remoteHost remotePort timeout waitRemoteClose} {
	global connected
    global portNr conf
    global state

	warning "trying to connect ($waitRemoteClose) to tcp!$remoteHost!$remotePort"
	after $timeout {set connected timeout}
    if {[catch {socket $remoteHost $remotePort} socket]} {
    	fatal "while connecting to tcp!$remoteHost!$remotePort: $socket" 1
    }
    set portNr [lindex [fconfigure $socket -sockname] 2]
    fileevent $socket w {set connected ok}
    vwait connected
    fileevent $socket w {}
    if {$connected == "timeout"} {
    	return -code error timeout
    } else {
	warning "connected to tcp!$remoteHost!$remotePort at $portNr"
	configure_fids $socket
	filevent_fids $socket $remoteHost $remotePort $waitRemoteClose
    }
}

proc getRemote {remoteHost remotePort sock} {
	fileevent $sock readable {}
	set read [gets $sock line]
	debug "getRemote tcp!$remoteHost!$remotePort read: $read"
	if {$read < 0} {
		if {[eof $sock]} {
			warning "Connection closed by client tcp!$remoteHost!$remotePort."
			cleanup_exit
		} elseif {[fblocked $sock]} {
			warning "getRemote tcp!$remoteHost!$remotePort blocked: $line"
		} else {
			warning "getRemote tcp!$remoteHost!$remotePort error?: $line"
			cleanup_exit
		}
	} else {
		output stdout $line
		debug "getRemote tcp!$remoteHost!$remotePort: $line"
	}
	fileevent $sock readable [list getRemote $remoteHost $remotePort $sock]
}
proc sendRemote {remoteHost remotePort sock waitRemoteClose} {
	fileevent stdin readable {}
	set read [gets stdin line]
	debug "sendRemote tcp!$remoteHost!$remotePort read: $read"
	if {$read < 0} {
		if {[eof stdin]} {
			# set line "A_QUIT"
			# catch {output $sock $line}
			debug "eof stdin"
			if {! $waitRemoteClose} {
				catch {close $sock}
				debug "sendRemote closed tcp!$remoteHost!$remotePort"
				cleanup_exit
			} else {
				return
			}
		} elseif {[fblocked stdin]} {
			warning "sendRemote tcp!$remoteHost!$remotePort blocked: $line"
		} else {
			warning "sendRemote tcp!$remoteHost!$remotePort error?: $line"
			cleanup_exit
		}
	} else {
		catch {output $sock $line}
		debug "sendRemote tcp!$remoteHost!$remotePort: $line"
	}
	fileevent stdin readable [list sendRemote $remoteHost $remotePort $sock $waitRemoteClose]
}

proc main_loop {sock waitRemoteClose argstring} {
	configure_fids $sock
	set remote [fconfigure $sock -peername]
	set remoteHost [lindex $remote 0]
	set remotePort [lindex $remote 2]
	puts $sock $argstring
	filevent_fids $sock $remoteHost $remotePort $waitRemoteClose
}

proc main {} {
	global argv
	global env

	set waitRemoteClose 0

	set reuse ""
	set mcast ""
	set title ""
	set key ""
	set files {}
	set exit 0

	set argstring ""
	
	if {[info exists env(TORXMCASTID)]} {
		set mcast "-m $env(TORXMCASTID)"
	}

	set args $argv
	
#	set joinedargs [join $argv ")(" ]
#	warning "command lin args: ($joinedargs)"

	while {[llength $args] > 0} {
		set a [lindex $args 0]
		set args [lrange $args 1 end]
		switch -- $a {
		    -r {
				set reuse "-r"
		  } -m {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
				} else {
					set mcast "-m [lindex $args 0]"
					set args [lrange $args 1 end]
				}
		  } -t {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
				} else {
					set title "-t {[lindex $args 0]}"
					set args [lrange $args 1 end]
				}
		  } -k {
				if {[llength $args] <= 0} {
		  			warning "missing argument for flag: $a"
				} else {
					set key "-k [lindex $args 0]"
					set args [lrange $args 1 end]
				}
		  } -exit {
		  		set exit 1
		  } -* {
		  		warning "unknown flag: $a"
		  } default {
		  		lappend files $a
		  }
		}
	}

	if {$exit || [llength $files] == 0} {
		set host [exec uname -n]
		set display [find_display $host]
		set user [find_user $host]
		set fname [mk_filename /tmp aniwait $user $display]
		# puts stderr "fname=$fname"

		set sock [try_connect $fname 0]

		set argstring "-args $reuse $mcast $title"

		if {[string compare $sock 0] != 0} {
			if {$exit} {
				puts $sock -exit
				exit
			}
			# puts stderr "sending $argstring"
			main_loop $sock $waitRemoteClose $argstring
		} elseif {! $exit} {
			exec aniwaitsrv &
			set i 0
			set maxtries 10
			while {$i < $maxtries} {
				if {$i == ($maxtries - 1)} {
					set sock [try_connect $fname 1]
				} else {
					set sock [try_connect $fname 0]
				}
				if {[string compare $sock 0] != 0} {
					main_loop $sock $waitRemoteClose $argstring
				}
				after 1000
				incr i
			}
		}
	} else {
		warning "usage: aniwait \[-r\] \[-m mcastid\] \[-t title\]"
		exit 1
	}
}

main
